"""Test cases for parsing list of positional and keyword arguments."""
import unittest

from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from typing import Union

from liquid.expression import Expression
from liquid.expression import FloatLiteral
from liquid.expression import IdentifierPathElement
from liquid.expression import Identifier
from liquid.expression import IntegerLiteral
from liquid.expression import StringLiteral
from liquid.expression import FALSE
from liquid.expression import TRUE
from liquid.expression import NIL

from liquid.expressions import parse_call_arguments
from liquid.expressions import parse_keyword_arguments
from liquid.expressions import parse_macro_arguments

from liquid.expressions.arguments.lex import tokenize
from liquid.expressions.arguments.parse import parse_equals_separated_arguments
from liquid.expressions.stream import TokenStream

from liquid.exceptions import Error
from liquid.exceptions import LiquidSyntaxError

Arguments = List[Tuple[Optional[str], Optional[Expression]]]


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    expression: str
    expect: Union[Arguments, Error]


class ParseKeywordArgumentsTestCase(unittest.TestCase):
    """Test cases for parsing keyword arguments."""

    def test_parse_keywords(self) -> None:
        """Test that we can parse a list of keyword arguments."""
        test_cases = [
            Case(
                description="empty expression",
                expression="",
                expect=[],
            ),
            Case(
                description="string literal value",
                expression="val: 'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="no whitespace",
                expression="val:'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="extra whitespace",
                expression="val :    'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="newline",
                expression="val:\n'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="leading comma",
                expression=", val: 'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="trailing comma",
                expression="val: 'hello',",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="multiple arguments",
                expression="a: 'hello', b: 'goodbye', c: 'you'",
                expect=[
                    ("a", StringLiteral("hello")),
                    ("b", StringLiteral("goodbye")),
                    ("c", StringLiteral("you")),
                ],
            ),
            Case(
                description="literal types",
                expression="a: 'hello', b: false, c: true, d: nil, e: 1, f: 1.1",
                expect=[
                    ("a", StringLiteral("hello")),
                    ("b", FALSE),
                    ("c", TRUE),
                    ("d", NIL),
                    ("e", IntegerLiteral(1)),
                    ("f", FloatLiteral(1.1)),
                ],
            ),
            Case(
                description="identifiers",
                expression=(
                    "a: title, b: product.title, "
                    "c: products[0], d: products['foo'].title"
                ),
                expect=[
                    ("a", Identifier([IdentifierPathElement("title")])),
                    (
                        "b",
                        Identifier(
                            [
                                IdentifierPathElement("product"),
                                IdentifierPathElement("title"),
                            ]
                        ),
                    ),
                    (
                        "c",
                        Identifier(
                            [
                                IdentifierPathElement("products"),
                                IdentifierPathElement(0),
                            ]
                        ),
                    ),
                    (
                        "d",
                        Identifier(
                            [
                                IdentifierPathElement("products"),
                                IdentifierPathElement("foo"),
                                IdentifierPathElement("title"),
                            ]
                        ),
                    ),
                ],
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                args = parse_keyword_arguments(case.expression)
                self.assertEqual(args, dict(case.expect))

    def test_parse_equals_separated_arguments(self) -> None:
        """Test that we can use `=` instead of `:` to separate key/value pairs."""
        expr = "a='hello', b=false, c=true, d=nil, e=1, f=1.1"
        expect = [
            ("a", StringLiteral("hello")),
            ("b", FALSE),
            ("c", TRUE),
            ("d", NIL),
            ("e", IntegerLiteral(1)),
            ("f", FloatLiteral(1.1)),
        ]
        args = dict(parse_equals_separated_arguments(TokenStream(tokenize(expr))))
        self.assertEqual(args, dict(expect))

    def test_missing_comma(self) -> None:
        """Test that we handle missing commas."""
        expr = "a: 'hello' b: 'goodbye'"
        with self.assertRaises(LiquidSyntaxError):
            parse_keyword_arguments(expr)

    def test_too_many_commas(self) -> None:
        """Test that we handle excess commas."""
        expr = "a: 'hello',, b: 'goodbye'"
        with self.assertRaises(LiquidSyntaxError):
            parse_keyword_arguments(expr)

    def test_missing_colon(self) -> None:
        """Test that we handle missing colons."""
        expr = "a 'hello'"
        with self.assertRaises(LiquidSyntaxError):
            parse_keyword_arguments(expr)

    def test_too_many_colon(self) -> None:
        """Test that we handle excess colons."""
        expr = "a:: 'hello'"
        with self.assertRaises(LiquidSyntaxError):
            parse_keyword_arguments(expr)


class ParseCallArgumentsTestCase(unittest.TestCase):
    """Test cases for parsing call-style argument lists."""

    def test_parse_call_arguments(self) -> None:
        """Test that we can parse a list of positional and/or keyword arguments."""
        test_cases = [
            Case(
                description="empty expression",
                expression="",
                expect=[],
            ),
            Case(
                description="string literal value",
                expression="val: 'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="no whitespace",
                expression="val:'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="extra whitespace",
                expression="val :    'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="multiple arguments",
                expression="a: 'hello', b: 'goodbye', c: 'you'",
                expect=[
                    ("a", StringLiteral("hello")),
                    ("b", StringLiteral("goodbye")),
                    ("c", StringLiteral("you")),
                ],
            ),
            Case(
                description="literal types",
                expression="a: 'hello', b: false, c: true, d: nil, e: 1, f: 1.1",
                expect=[
                    ("a", StringLiteral("hello")),
                    ("b", FALSE),
                    ("c", TRUE),
                    ("d", NIL),
                    ("e", IntegerLiteral(1)),
                    ("f", FloatLiteral(1.1)),
                ],
            ),
            Case(
                description="identifiers",
                expression=(
                    "a: title, b: product.title, c: products[0], d: products[0].title"
                ),
                expect=[
                    ("a", Identifier([IdentifierPathElement("title")])),
                    (
                        "b",
                        Identifier(
                            [
                                IdentifierPathElement("product"),
                                IdentifierPathElement("title"),
                            ]
                        ),
                    ),
                    (
                        "c",
                        Identifier(
                            [
                                IdentifierPathElement("products"),
                                IdentifierPathElement(0),
                            ]
                        ),
                    ),
                    (
                        "d",
                        Identifier(
                            [
                                IdentifierPathElement("products"),
                                IdentifierPathElement(0),
                                IdentifierPathElement("title"),
                            ]
                        ),
                    ),
                ],
            ),
            Case(
                description="positional argument",
                expression="'hello'",
                expect=[(None, StringLiteral("hello"))],
            ),
            Case(
                description="multiple positional arguments",
                expression="'hello', 'you'",
                expect=[(None, StringLiteral("hello")), (None, StringLiteral("you"))],
            ),
            Case(
                description="positional no whitespace",
                expression="'hello','you'",
                expect=[(None, StringLiteral("hello")), (None, StringLiteral("you"))],
            ),
            Case(
                description="positional extra whitespace",
                expression="'hello'  ,    'you'",
                expect=[(None, StringLiteral("hello")), (None, StringLiteral("you"))],
            ),
            Case(
                description="positional argument followed by keyword",
                expression="'hello', greeting: 'you'",
                expect=[
                    (None, StringLiteral("hello")),
                    ("greeting", StringLiteral("you")),
                ],
            ),
            Case(
                description="keyword argument followed by positional",
                expression="greeting: 'you', 'hello'",
                expect=[
                    ("greeting", StringLiteral("you")),
                    (None, StringLiteral("hello")),
                ],
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                expr = "'func' " + case.expression  # pretend macro name
                _, args = parse_call_arguments(expr)
                self.assertEqual(args, case.expect)

    def test_missing_comma(self) -> None:
        """Test that we handle missing commas."""
        expr = "'func' a: 'hello' b: 'goodbye'"
        with self.assertRaises(LiquidSyntaxError):
            parse_call_arguments(expr)

    def test_too_many_commas(self) -> None:
        """Test that we handle excess commas."""
        expr = "'func' a: 'hello',, b: 'goodbye'"
        with self.assertRaises(LiquidSyntaxError):
            parse_call_arguments(expr)


class ParseMacroArgumentsTestCase(unittest.TestCase):
    """Test cases for parsing macro-style argument lists."""

    def test_parse_call_arguments(self) -> None:
        """Test that we can parse a list of positional and/or keyword arguments."""
        test_cases = [
            Case(
                description="empty expression",
                expression="",
                expect=[],
            ),
            Case(
                description="string literal value",
                expression="val: 'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="no whitespace",
                expression="val:'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="extra whitespace",
                expression="val :    'hello'",
                expect=[("val", StringLiteral("hello"))],
            ),
            Case(
                description="multiple arguments",
                expression="a: 'hello', b: 'goodbye', c: 'you'",
                expect=[
                    ("a", StringLiteral("hello")),
                    ("b", StringLiteral("goodbye")),
                    ("c", StringLiteral("you")),
                ],
            ),
            Case(
                description="literal types",
                expression="a: 'hello', b: false, c: true, d: nil, e: 1, f: 1.1",
                expect=[
                    ("a", StringLiteral("hello")),
                    ("b", FALSE),
                    ("c", TRUE),
                    ("d", NIL),
                    ("e", IntegerLiteral(1)),
                    ("f", FloatLiteral(1.1)),
                ],
            ),
            Case(
                description="identifiers",
                expression=(
                    "a: title, b: product.title, c: products[0], d: products[0].title"
                ),
                expect=[
                    ("a", Identifier([IdentifierPathElement("title")])),
                    (
                        "b",
                        Identifier(
                            [
                                IdentifierPathElement("product"),
                                IdentifierPathElement("title"),
                            ]
                        ),
                    ),
                    (
                        "c",
                        Identifier(
                            [
                                IdentifierPathElement("products"),
                                IdentifierPathElement(0),
                            ]
                        ),
                    ),
                    (
                        "d",
                        Identifier(
                            [
                                IdentifierPathElement("products"),
                                IdentifierPathElement(0),
                                IdentifierPathElement("title"),
                            ]
                        ),
                    ),
                ],
            ),
            Case(
                description="identifier without a default value",
                expression="name",
                expect=[("name", NIL)],
            ),
            Case(
                description="multiple identifiers without a default value",
                expression="name, greeting",
                expect=[("name", NIL), ("greeting", NIL)],
            ),
            Case(
                description=(
                    "identifier without default followed by identifier with a default"
                ),
                expression="name, greeting: 'hello'",
                expect=[("name", NIL), ("greeting", StringLiteral("hello"))],
            ),
            Case(
                description=(
                    "identifier with a default followed by identifier without default"
                ),
                expression="greeting: 'hello', name",
                expect=[("greeting", StringLiteral("hello")), ("name", NIL)],
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                expr = "'func' " + case.expression  # pretend macro name
                name, args = parse_macro_arguments(expr)
                self.assertEqual(name, "func")
                self.assertEqual(args, case.expect)

    def test_missing_comma(self) -> None:
        """Test that we handle missing commas."""
        expr = "'func' a: 'hello' b: 'goodbye'"
        with self.assertRaises(LiquidSyntaxError):
            parse_macro_arguments(expr)

    def test_too_many_commas(self) -> None:
        """Test that we handle excess commas."""
        expr = "'func' a: 'hello',, b: 'goodbye'"
        with self.assertRaises(LiquidSyntaxError):
            parse_macro_arguments(expr)



"""Golden tests cases for testing liquid's built-in `sort_natural` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="array of strings",
        template=r"{{ a | sort_natural | join: '#' }}",
        expect="a#A#b#B#C",
        globals={"a": ["b", "a", "C", "B", "A"]},
    ),
    Case(
        description="array of strings with a nul",
        template=(
            r"{% assign x = a | sort_natural %}"
            r"{% for i in x %}"
            r"{{ i }}"
            r"{% unless forloop.last %}#{% endunless %}"
            r"{% endfor %}"
        ),
        expect="a#A#b#B#C#",
        globals={"a": ["b", "a", None, "C", "B", "A"]},
    ),
    Case(
        description="array of objects with a key",
        template=(
            r"{% assign x = a | sort_natural: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,bar)(title,Baz)(title,foo)",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}]},
    ),
    Case(
        description="array of objects with a key gets stringified",
        template=(
            r"{% assign x = a | sort_natural: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,1111)(title,87)(title,9)",
        globals={"a": [{"title": 9}, {"title": 1111}, {"title": 87}]},
    ),
    Case(
        description="array of objects with a missing key",
        template=(
            r"{% assign x = a | sort_natural: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,bar)(title,foo)(heading,Baz)",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"heading": "Baz"}]},
    ),
    Case(
        description="argument is undefined",
        template=(
            r"{% assign x = a | sort_natural: nosuchthing %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,bar)(title,Baz)(title,foo)",
        globals={"a": [{"title": "foo"}, {"title": "bar"}, {"title": "Baz"}]},
    ),
    Case(
        description="empty array",
        template=(
            r"{% assign x = a | sort_natural %}"
            r"{% for i in x %}"
            r"{{ i }}"
            r"{% unless forloop.last %}#{% endunless %}"
            r"{% endfor %}"
        ),
        expect="",
        globals={"a": []},
    ),
    Case(
        description="left value is not an array",
        template=r"{{ a | sort_natural }}",
        expect="123",
        globals={"a": 123},
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | sort_natural }}",
        expect="",
    ),
    Case(
        description="incompatible types",
        template=r"{{ a | sort_natural }}",
        expect="14{}",
        globals={"a": [{}, 1, "4"]},
    ),
]



"""Golden tests cases for testing liquid's built-in `join` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="range literal join filter left value",
        template=r"{{ (1..3) | join: '#' }}",
        expect="1#2#3",
    ),
    Case(
        description="joining a string is a noop",
        template=r"{{ 'a,b' | join: '#' }}",
        expect="a,b",
    ),
    Case(
        description="joining an int is a noop",
        template=r"{{ 123 | join: '#' }}",
        expect="123",
    ),
    Case(
        description="join an array of strings",
        template=r"{{ arr | join: '#' }}",
        expect="a#b",
        globals={"arr": ["a", "b"]},
    ),
    Case(
        description="join an array of integers",
        template=r"{{ arr | join: '#' }}",
        expect="1#2",
        globals={"arr": [1, 2]},
    ),
    Case(
        description="missing argument defaults to a space",
        template=r"{{ arr | join }}",
        expect="a b",
        globals={"arr": ["a", "b"]},
    ),
    Case(
        description="argument is not a string",
        template=r"{{ arr | join: 5 }}",
        expect="a5b",
        globals={"arr": ["a", "b"]},
    ),
    Case(
        description="left value contains non string",
        template=r"{{ arr | join: '#' }}",
        expect="a#b#1",
        globals={"arr": ["a", "b", 1]},
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | join: '#' }}",
        expect="",
    ),
    Case(
        description="undefined argument",
        template=r"{{ arr | join: nosuchthing }}",
        expect="ab",
        globals={"arr": ["a", "b"]},
    ),
    Case(
        description="too many arguments",
        template=r"{{ arr | join: '#', 42 }}",
        expect="",
        globals={"arr": ["a", "b"]},
        error=True,
    ),
]



"""Golden tests cases for testing liquid's `increment` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="named counter",
        template=r"{% increment foo %} {% increment foo %} {% increment foo %}",
        expect="0 1 2",
    ),
    Case(
        description="incrementing counter renders before incrementing",
        template=r"{% increment foo %} {{ foo }}",
        expect="0 1",
    ),
    Case(
        description="multiple named counters",
        template=(
            r"{% increment foo %} "
            r"{% increment bar %} "
            r"{% increment foo %} "
            r"{% increment bar %}"
        ),
        expect="0 0 1 1",
    ),
    Case(
        description="assign and increment",
        template=(
            r"{% assign foo = 5 %}"
            r"{{ foo }} "
            r"{% increment foo %} "
            r"{% increment foo %} "
            r"{{ foo }}"
        ),
        expect="5 0 1 5",
    ),
    Case(
        description="named counters are in scope for subsequent expressions",
        template=(
            r"{% increment foo %} "
            r"{% increment foo %} "
            r"{% if foo > 0 %}"
            r"{{ foo }}"
            r"{% endif %}"
        ),
        expect="0 1 2",
    ),
]



"""Liquid template parser.

As of Python Liquid 1.2.0 we are transitioning away from the expression parser defined
here. Instead opting for separate, specialized parsers (and lexers) for each of the
three built-in expression types.

The `ExpressionParser` class will become depreciated as we approach version 2.0 and then
removed. At which time this module will be reserved for parsing templates (not
expressions), possibly with some explicit re-exports of the expression parsers found in
`liquid.expressions`.

Developers of custom tags are encouraged to use or follow the example of:

- `liquid.expressions.parse_boolean_expression`
- `liquid.expressions.parse_filtered_expression`
- `liquid.expressions.parse_loop_expression`

And make use of `liquid.expressions.common` and `liquid.expressions.TokenStream`.
"""

from __future__ import annotations

from enum import IntEnum, auto
from functools import lru_cache

from typing import Tuple
from typing import List
from typing import Union
from typing import Optional
from typing import NamedTuple
from typing import Container
from typing import TYPE_CHECKING

from liquid import ast

from liquid import expression
from liquid.expression import IdentifierPathElement

from liquid.lex import tokenize_boolean_expression
from liquid.lex import tokenize_loop_expression
from liquid.lex import tokenize_filtered_expression

from liquid.limits import to_int
from liquid.stream import TokenStream

from liquid.exceptions import LiquidSyntaxError
from liquid.exceptions import Error

from liquid.token import TOKEN_ILLEGAL
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_STATEMENT
from liquid.token import TOKEN_LITERAL
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_FLOAT
from liquid.token import TOKEN_EMPTY
from liquid.token import TOKEN_NIL
from liquid.token import TOKEN_NULL
from liquid.token import TOKEN_BLANK
from liquid.token import TOKEN_NEGATIVE
from liquid.token import TOKEN_TRUE
from liquid.token import TOKEN_FALSE
from liquid.token import TOKEN_CONTAINS
from liquid.token import TOKEN_IN
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_LIMIT
from liquid.token import TOKEN_OFFSET
from liquid.token import TOKEN_REVERSED
from liquid.token import TOKEN_CONTINUE
from liquid.token import TOKEN_COLS
from liquid.token import TOKEN_PIPE
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_DOT
from liquid.token import TOKEN_LBRACKET
from liquid.token import TOKEN_RBRACKET
from liquid.token import TOKEN_ASSIGN
from liquid.token import TOKEN_AND
from liquid.token import TOKEN_OR
from liquid.token import TOKEN_EQ
from liquid.token import TOKEN_NE
from liquid.token import TOKEN_LG
from liquid.token import TOKEN_LT
from liquid.token import TOKEN_GT
from liquid.token import TOKEN_LE
from liquid.token import TOKEN_GE
from liquid.token import reverse_operators
from liquid.token import Token

if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment


class Parser:
    """A Liquid template parser. Create a parse tree from a stream of tokens."""

    __slots__ = ("tags", "illegal", "literal", "statement", "env")

    def __init__(self, env: Environment):
        self.tags = env.tags
        self.env = env

        self.illegal = self.tags[TOKEN_ILLEGAL]
        self.literal = self.tags[TOKEN_LITERAL]
        self.statement = self.tags[TOKEN_STATEMENT]

    def parse(self, stream: TokenStream) -> ast.ParseTree:
        """Parse the given stream of tokens into a tree."""
        root = ast.ParseTree()
        statements = root.statements

        while stream.current.type != TOKEN_EOF:
            try:
                statements.append(self.parse_statement(stream))
            except Error as err:
                self.env.error(err, linenum=stream.current.linenum)

            stream.next_token()

        return root

    def parse_statement(self, stream: TokenStream) -> ast.Node:
        """Parse a node from a stream of tokens."""
        if stream.current.type == TOKEN_STATEMENT:
            node = self.statement.get_node(stream)
        elif stream.current.type == TOKEN_TAG:
            tag = self.tags.get(stream.current.value, self.illegal)

            if tag.block:
                stream.balancing_stack.append(tag.end)

            node = tag.get_node(stream)

            # Tag parse functions can choose to return an IllegalNode.
            if isinstance(node, ast.IllegalNode):
                raise LiquidSyntaxError(
                    f"unexpected tag '{node.token().value}'",
                    linenum=node.token().linenum,
                )

            if tag.block:
                popped = stream.balancing_stack.pop()
                if stream.current.value != popped:
                    raise LiquidSyntaxError(
                        f"expected {popped}, found {stream.current.value}",
                        linenum=stream.current.linenum,
                    )
        else:
            node = self.literal.get_node(stream)

        return node

    def parse_block(self, stream: TokenStream, end: Container[str]) -> ast.BlockNode:
        """Parse multiple nodes from a stream of tokens. Stop parsing nodes when we
        find a token in `end` or we reach the end of the stream."""
        block = ast.BlockNode(stream.current)
        statements = block.statements

        while stream.current.type != TOKEN_EOF:
            if stream.current.type == TOKEN_TAG and stream.current.value in end:
                break
            stmt = self.parse_statement(stream)
            statements.append(stmt)
            # Detect output nodes in any of this block's children. This is used by
            # some tags to automatically suppress whitespace when no other output is
            # present.
            if stmt.force_output or getattr(stmt, "forced_output", False):
                block.forced_output = True
            stream.next_token()

        return block


def eat_block(stream: TokenStream, end: Container[str]) -> None:
    """Advance the stream pointer past the next end tag.

    This is used to discard blocks after a syntax error is found, in the hope
    that we can continue to parse more of the stream after the offending block.
    """
    while stream.current.type != TOKEN_EOF:
        if stream.current.type == TOKEN_TAG and stream.current.value in end:
            break
        stream.next_token()


@lru_cache(maxsize=128)
def get_parser(env: Environment) -> Parser:
    """Return a template parser for the given environment."""
    return Parser(env)


# NOTE: Everything from here on is flagged for future depreciation. See the comments at
# the top of this module.


def _expect(tok: Token, typ: str, value: Optional[str] = None) -> None:
    if tok.type != typ or (value is not None and tok.value != value):
        _typ = reverse_operators.get(tok.type, tok.type)
        _expected_typ = reverse_operators.get(typ, typ)
        if value is not None:
            msg = (
                f"expected {_expected_typ} with value '{value}', "
                f"found {_typ} with value '{tok.value}'"
            )
        else:
            msg = f"expected '{_expected_typ}', found '{_typ}'"
        raise LiquidSyntaxError(msg, linenum=tok.linenum)


def expect(stream: TokenStream, typ: str, value: Optional[str] = None) -> None:
    """Raise an exception if the current token in the stream does not match the given
    type and value."""
    _expect(stream.current, typ, value)


def expect_peek(stream: TokenStream, typ: str, value: Optional[str] = None) -> None:
    """Raise an exception if the next token in the stream does not match the given
    type and value."""
    _expect(stream.peek, typ, value)


def is_end_tag(tok: Token) -> bool:
    """Return `True` if the current token looks like an end tag."""
    return tok.value.startswith("end")


class Precedence(IntEnum):
    """Operator precedence."""

    LOWEST = 1
    LOGICALRIGHT = 49
    LOGICAL = 50
    RELATIONAL = auto()
    MEMBERSHIP = auto()
    PREFIX = auto()


PRECEDENCES = {
    TOKEN_IDENTIFIER: Precedence.LOWEST,
    TOKEN_EQ: Precedence.RELATIONAL,
    TOKEN_LT: Precedence.RELATIONAL,
    TOKEN_GT: Precedence.RELATIONAL,
    TOKEN_NE: Precedence.RELATIONAL,
    TOKEN_LG: Precedence.RELATIONAL,
    TOKEN_LE: Precedence.RELATIONAL,
    TOKEN_GE: Precedence.RELATIONAL,
    TOKEN_CONTAINS: Precedence.MEMBERSHIP,
    TOKEN_AND: Precedence.LOGICAL,
    TOKEN_OR: Precedence.LOGICAL,
}

RIGHT_ASSOCIATIVE = frozenset((TOKEN_AND, TOKEN_OR))

IDENTIFIER_TOKENS = frozenset(
    (TOKEN_IDENTIFIER, TOKEN_INTEGER, TOKEN_DOT, TOKEN_LBRACKET, TOKEN_STRING)
)

LOOP_TOKENS = frozenset(
    (
        TOKEN_LIMIT,
        TOKEN_OFFSET,
        TOKEN_COLS,
        TOKEN_REVERSED,
    )
)

FILTER_END_TOKENS = frozenset((TOKEN_PIPE, TOKEN_EOF))


def parse_boolean(stream: TokenStream) -> expression.Boolean:
    """Read a boolean literal (true or false) from the token stream."""
    if stream.current.type == TOKEN_TRUE:
        return expression.TRUE
    return expression.FALSE


def parse_nil(_: TokenStream) -> expression.Nil:
    """Read a 'nil' keyword from the token stream."""
    return expression.NIL


def parse_empty(_: TokenStream) -> expression.Empty:
    """Read a 'empty' keyword from the token stream."""
    return expression.EMPTY


def parse_blank(_: TokenStream) -> expression.Blank:
    """Read a 'blank' keyword from the token stream."""
    return expression.BLANK


def parse_string_literal(stream: TokenStream) -> expression.StringLiteral:
    """Read a string from the token stream."""
    return expression.StringLiteral(value=stream.current.value)


def parse_integer_literal(stream: TokenStream) -> expression.IntegerLiteral:
    """Read an integer from the token stream."""
    return expression.IntegerLiteral(value=to_int(stream.current.value))


def parse_float_literal(stream: TokenStream) -> expression.FloatLiteral:
    """Read a float from the token stream."""
    return expression.FloatLiteral(value=float(stream.current.value))


def parse_range_literal(stream: TokenStream) -> expression.RangeLiteral:
    """Read a range literal from the token stream."""
    # Start of a range expression (<int or id>..<int or id>)
    expect(stream, TOKEN_LPAREN)
    stream.next_token()
    start = parse_range_argument(stream)

    expect_peek(stream, TOKEN_RANGE)
    stream.next_token()
    stream.next_token()  # Eat TOKEN_RANGE

    stop = parse_range_argument(stream)
    expect_peek(stream, TOKEN_RPAREN)

    assert isinstance(
        start,
        (expression.Identifier, expression.IntegerLiteral, expression.FloatLiteral),
    )
    assert isinstance(
        stop,
        (expression.Identifier, expression.IntegerLiteral, expression.FloatLiteral),
    )

    expr = expression.RangeLiteral(start, stop)
    stream.next_token()
    return expr


def parse_identifier(stream: TokenStream) -> expression.Identifier:
    """Read an identifier from the token stream.

    <ident>.<ident>
    <ident>["<ident>"]
    <ident>["<ident>"].<ident>
    <ident>[<ident --> int/str>]
    <ident>[<ident>.<ident --> int/str>]
    <ident>[<int>]
    <ident>[<int>].<ident>
    """
    path: expression.IdentifierPath = []

    while stream.current.type in IDENTIFIER_TOKENS:
        if stream.current.type == TOKEN_IDENTIFIER:
            path.append(IdentifierPathElement(stream.current.value))

        elif stream.current.type == TOKEN_INTEGER:
            path.append(IdentifierPathElement(to_int(stream.current.value)))

        elif stream.current.type == TOKEN_LBRACKET:
            stream.next_token()  # Eat open bracket

            if stream.current.type == TOKEN_STRING:
                path.append(IdentifierPathElement(stream.current.value))
            elif stream.current.type == TOKEN_NEGATIVE:
                expect_peek(stream, TOKEN_INTEGER)
                stream.next_token()
                path.append(IdentifierPathElement(-to_int(stream.current.value)))
            elif stream.current.type == TOKEN_INTEGER:
                path.append(IdentifierPathElement(to_int(stream.current.value)))

            elif stream.current.type == TOKEN_IDENTIFIER:
                # Recursive call to parse_identifier. If it's not a string or
                # integer, anything inside a pair of square brackets could be
                # another identifier that resolves to a string or integer.
                path.append(parse_identifier(stream))
            else:
                raise LiquidSyntaxError(
                    f"invalid identifier, found {stream.current.type}"
                )

            expect_peek(stream, TOKEN_RBRACKET)
            stream.next_token()  # Eat close bracket

        elif stream.current.type == TOKEN_DOT:
            pass
        else:
            raise LiquidSyntaxError(f"invalid identifier, found {stream.current.type}")

        stream.next_token()

    stream.push(stream.current)
    return expression.Identifier(path)


def parse_string_or_identifier(
    stream: TokenStream,
    linenum: Optional[int] = None,
) -> expression.Expression:
    """Parse an expression from a stream of tokens. If the stream is not at a string or
    identifier expression, raise a syntax error.
    """
    if stream.current.type == TOKEN_IDENTIFIER:
        expr: expression.Expression = parse_identifier(stream)
    elif stream.current.type == TOKEN_STRING:
        expr = parse_string_literal(stream)
    else:
        _typ = reverse_operators.get(stream.current.type, stream.current.type)
        msg = f"expected {TOKEN_IDENTIFIER} or {TOKEN_STRING}, found {_typ}"
        raise LiquidSyntaxError(msg, linenum=linenum or stream.current.linenum)

    return expr


def parse_unchained_identifier(
    stream: TokenStream,
    linenum: Optional[int] = None,
) -> expression.Identifier:
    """Parse an identifier from a stream of tokens. If the stream is not at an
    identifier or the identifier is chained, raise a syntax error."""
    expect(stream, TOKEN_IDENTIFIER)
    tok = stream.current
    ident = parse_identifier(stream)

    if len(ident.path) != 1:
        raise LiquidSyntaxError(
            f"invalid identifier '{ident}'", linenum=linenum or tok.linenum
        )

    return ident


# NOTE: These names are misleading. The args and options in question can apply to any
# iterable in a loop expression, not just ranges.

RangeArg = Union[
    expression.Identifier,
    expression.IntegerLiteral,
    expression.FloatLiteral,
    expression.Continue,
    expression.Nil,
]


class RangeOption(NamedTuple):
    """A token and value representing a keyword argument in a loop expression."""

    tok: Token
    arg: RangeArg


def parse_range_argument(stream: TokenStream) -> RangeArg:
    """Parse a loop argument value from a stream of tokens. If the stream is not at a
    valid token for a loop argument value, raise a syntax error."""
    if stream.current.type == TOKEN_IDENTIFIER:
        arg: RangeArg = parse_identifier(stream)
    elif stream.current.type == TOKEN_INTEGER:
        arg = parse_integer_literal(stream)
    elif stream.current.type == TOKEN_FLOAT:
        arg = parse_float_literal(stream)
    elif stream.current.type == TOKEN_CONTINUE:
        arg = expression.CONTINUE
    else:
        raise LiquidSyntaxError(
            "invalid range expression, expected an integer, "
            f"found a {stream.current.type}"
        )
    return arg


def parse_range_option(stream: TokenStream) -> RangeOption:
    """Parse a loop keyword argument from a stream of tokens. If the stream is not at a
    valid loop keyword argument, raise a syntax error."""
    tok = stream.current
    if stream.current.type in (TOKEN_LIMIT, TOKEN_OFFSET, TOKEN_COLS):
        expect_peek(stream, TOKEN_COLON)
        stream.next_token()
        stream.next_token()  # Eat COLON
        arg = parse_range_argument(stream)
        stream.next_token()
        opt = RangeOption(tok=tok, arg=arg)
    else:
        expect(stream, TOKEN_REVERSED)
        stream.next_token()  # Eat REVERSED
        opt = RangeOption(tok=tok, arg=expression.NIL)

    return opt


def parse_loop_expression_value(value: str) -> expression.LoopExpression:
    """Parse the given string as a loop expression."""
    return parse_loop_expression(TokenStream(tokenize_loop_expression(value)))


def parse_loop_expression(stream: TokenStream) -> expression.LoopExpression:
    """Parse a liquid loop expression, as one might find in a `for` tag.

    If any of the optional parameters are duplicated, the last (right most) is
    used.
    """
    expect(stream, TOKEN_IDENTIFIER)
    expect_peek(stream, TOKEN_IN)

    # Loop variable name
    name = stream.current.value

    stream.next_token()
    stream.next_token()  # Eat TOKEN_IN

    if stream.current.type == TOKEN_IDENTIFIER:
        # Identifier should resolve to an array or hash/map at render time.
        iterable: expression.LoopIterable = parse_identifier(stream)
        stream.next_token()
    elif stream.current.type == TOKEN_LPAREN:
        iterable = parse_range_literal(stream)
        stream.next_token()
    else:
        raise LiquidSyntaxError("invalid loop expression")

    expr = expression.LoopExpression(name=name, iterable=iterable)

    # Optional range modifiers can appear in any order.
    while stream.current.type in LOOP_TOKENS:
        opt = parse_range_option(stream)

        if opt.tok.type == TOKEN_REVERSED:
            expr.reversed = True
        elif opt.tok.type == TOKEN_LIMIT:
            expr.limit = opt.arg
        elif opt.tok.type == TOKEN_OFFSET:
            expr.offset = opt.arg
        elif opt.tok.type == TOKEN_COLS:
            expr.cols = opt.arg

    return expr


class ExpressionParser:
    """The standard expression parser.

    Custom, non "standard" tags can subclass `ExpressionParser` to change expression
    parsing behavior. Such as adding a logical `NOT` operator.
    """

    END_EXPRESSION: Tuple[str, ...] = (
        TOKEN_PIPE,
        TOKEN_COMMA,
        TOKEN_EOF,
    )

    def __init__(self) -> None:
        self.precedences = PRECEDENCES.copy()
        self.right_associative = RIGHT_ASSOCIATIVE.copy()

        self.prefix_funcs = {
            TOKEN_FALSE: parse_boolean,
            TOKEN_TRUE: parse_boolean,
            TOKEN_NIL: parse_nil,
            TOKEN_NULL: parse_nil,
            TOKEN_EMPTY: parse_empty,
            TOKEN_BLANK: parse_blank,
            TOKEN_NEGATIVE: self.parse_prefix_expression,
            TOKEN_STRING: parse_string_literal,
            TOKEN_INTEGER: parse_integer_literal,
            TOKEN_FLOAT: parse_float_literal,
            TOKEN_IDENTIFIER: parse_identifier,
            TOKEN_LPAREN: parse_range_literal,
        }

        self.infix_funcs = {
            TOKEN_EQ: self.parse_infix_expression,
            TOKEN_OR: self.parse_infix_expression,
            TOKEN_AND: self.parse_infix_expression,
            TOKEN_LT: self.parse_infix_expression,
            TOKEN_GT: self.parse_infix_expression,
            TOKEN_NE: self.parse_infix_expression,
            TOKEN_LG: self.parse_infix_expression,
            TOKEN_LE: self.parse_infix_expression,
            TOKEN_GE: self.parse_infix_expression,
            TOKEN_CONTAINS: self.parse_infix_expression,
        }

    def peek_precedence(self, stream: TokenStream) -> Precedence:
        """Return the precedence of the next token in the stream."""
        try:
            precedence = self.precedences[stream.peek.type]
        except KeyError as err:
            raise LiquidSyntaxError(f"unknown operator '{stream.peek.value}'") from err
        return precedence

    def current_precedence(self, stream: TokenStream) -> Precedence:
        """Return the precedence of the current token in the stream."""
        precedence = self.precedences.get(stream.current.type, Precedence.LOWEST)
        if stream.current.type in self.right_associative:
            precedence = Precedence(precedence - 1)
        return Precedence(precedence)

    def parse_filters(self, stream: TokenStream) -> List[expression.Filter]:
        """Keep reading filters from the token stream until end of expression."""
        filters = []
        while stream.current.type != TOKEN_EOF:
            filters.append(self.parse_filter(stream))
        return filters

    def parse_filter(self, stream: TokenStream) -> expression.Filter:
        """Read a single filter with optional arguments from the token stream."""
        # Eat the pipe token
        expect(stream, TOKEN_PIPE)
        stream.next_token()

        # Read the filter's name
        expect(stream, TOKEN_IDENTIFIER)
        filter_name = stream.current.value
        stream.next_token()

        #
        args = []
        kwargs = {}

        if stream.current.type == TOKEN_COLON:
            stream.next_token()

            while stream.current.type not in FILTER_END_TOKENS:
                if stream.peek.type == TOKEN_COLON:
                    # A named parameter
                    param_name = stream.current.value
                    stream.next_token()

                    # Eat colon
                    stream.next_token()

                    param_value = self.parse_expression(stream)
                    kwargs[param_name] = param_value
                else:
                    # A positional argument/parameter
                    args.append(self.parse_expression(stream))
                stream.next_token()

                if stream.current.type == TOKEN_COMMA:
                    stream.next_token()

        return expression.Filter(filter_name, args, kwargs)

    def parse_prefix_expression(
        self, stream: TokenStream
    ) -> expression.PrefixExpression:
        """Parse a prefix expression from a stream of tokens."""
        tok = stream.current
        stream.next_token()

        exp = expression.PrefixExpression(
            tok.value,
            right=self.parse_expression(stream, precedence=Precedence.PREFIX),
        )

        return exp

    def parse_infix_expression(
        self,
        stream: TokenStream,
        left: expression.Expression,
    ) -> expression.InfixExpression:
        """Parse an infix expression from a stream of tokens."""
        tok = stream.current
        precedence = self.current_precedence(stream)
        stream.next_token()

        exp = expression.InfixExpression(
            left=left,
            operator=tok.value,
            right=self.parse_expression(stream, precedence),
        )
        return exp

    def parse_expression(
        self,
        stream: TokenStream,
        precedence: Precedence = Precedence.LOWEST,
    ) -> expression.Expression:
        """Parse a literal or logical expression."""
        end_expression = self.END_EXPRESSION
        infix_funcs = self.infix_funcs

        prefix = self.prefix_funcs.get(stream.current.type)
        if prefix is None:
            operator = reverse_operators.get(stream.current.type, stream.current.value)
            raise LiquidSyntaxError(f"unknown prefix operator '{operator}'")

        left = prefix(stream)

        while (
            stream.peek.type not in end_expression
            and precedence < self.peek_precedence(stream)
        ):
            infix = infix_funcs.get(stream.peek.type)
            if infix is None:
                return left

            stream.next_token()
            left = infix(stream, left)

        return left

    def parse_filtered_expression_value(
        self, value: str
    ) -> expression.FilteredExpression:
        """Parse the given value as a filtered expression."""
        return self.parse_filtered_expression(
            TokenStream(tokenize_filtered_expression(value))
        )

    def parse_filtered_expression(
        self, stream: TokenStream
    ) -> expression.FilteredExpression:
        """Parse a statement (output) expression with optional filters.

        A statement expression is assumed to start with exactly one identifier, followed
        by zero or more filters.
        """
        expr = self.parse_expression(stream)
        stream.next_token()
        filters = self.parse_filters(stream)
        return expression.FilteredExpression(expr, filters)

    def parse_boolean_expression_value(self, value: str) -> expression.Expression:
        """Parse the given string as a boolean expression."""
        return self.parse_boolean_expression(
            TokenStream(tokenize_boolean_expression(value))
        )

    def parse_boolean_expression(self, stream: TokenStream) -> expression.Expression:
        """Parse a liquid expression that evaluates to a Boolean value.

        This is primarily used by control flow tags like `if`, `unless` and `case`/
        `when`.

        Logical operators (`and` and `or`) are right associative. They are evaluated
        from right to left. The use of parentheses to control operator precedence is
        not allowed.
        """
        if stream.current.type == TOKEN_EOF:
            # Empty expression.
            return expression.NIL
        return expression.BooleanExpression(expression=self.parse_expression(stream))

    def parse_assignment_expression(
        self, stream: TokenStream
    ) -> expression.AssignmentExpression:
        """Parse a liquid assignment expression, as one might find in an `assign` tag.

        This is essentially the same as a parse_filtered_expression, but with
        an additional name to bind the expression result to.
        """
        expect(stream, TOKEN_IDENTIFIER)
        expect_peek(stream, TOKEN_ASSIGN)

        name = parse_unchained_identifier(stream)

        stream.next_token()
        stream.next_token()  # Eat ASSIGN TOKEN

        expr = self.parse_filtered_expression(stream)

        return expression.AssignmentExpression(name=str(name), expression=expr)


# For backwards compatibility.
STANDARD_EXPRESSION_PARSER = ExpressionParser()
peek_precedence = STANDARD_EXPRESSION_PARSER.peek_precedence
current_precedence = STANDARD_EXPRESSION_PARSER.current_precedence
parse_filters = STANDARD_EXPRESSION_PARSER.parse_filters
parse_filter = STANDARD_EXPRESSION_PARSER.parse_filter
parse_prefix_expression = STANDARD_EXPRESSION_PARSER.parse_prefix_expression
parse_infix_expression = STANDARD_EXPRESSION_PARSER.parse_infix_expression
parse_expression = STANDARD_EXPRESSION_PARSER.parse_expression
parse_filtered_expression = STANDARD_EXPRESSION_PARSER.parse_filtered_expression
parse_filtered_expression_value = (
    STANDARD_EXPRESSION_PARSER.parse_filtered_expression_value
)
parse_boolean_expression = STANDARD_EXPRESSION_PARSER.parse_boolean_expression
parse_boolean_expression_value = (
    STANDARD_EXPRESSION_PARSER.parse_boolean_expression_value
)
parse_assignment_expression = STANDARD_EXPRESSION_PARSER.parse_assignment_expression


@lru_cache(maxsize=128)
def get_expression_parser(_: Environment) -> ExpressionParser:
    """Return an expression parser for the given environment."""
    # Future proofing.
    return STANDARD_EXPRESSION_PARSER



"""Golden tests cases for testing liquid's built-in `url_decode` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="some special URL characters",
        template=r'{{ "email+address+is+bob%40example.com%21" | url_decode }}',
        expect="email address is bob@example.com!",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | url_decode }}",
        expect="5",
    ),
    Case(
        description="unexpected argument",
        template=r'{{ "hello" | url_decode: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | url_decode }}",
        expect="",
    ),
]



"""Liquid template and expression parsing tests."""
# pylint: disable=missing-class-docstring missing-function-docstring

import unittest

from typing import Iterable
from typing import List
from typing import NamedTuple
from typing import Union

from liquid.environment import Environment
from liquid.mode import Mode

from liquid.builtin import literal
from liquid.builtin import statement
from liquid.builtin import if_tag
from liquid.builtin import comment_tag
from liquid.builtin import unless_tag
from liquid.builtin import case_tag
from liquid.builtin import for_tag
from liquid.builtin import tablerow_tag
from liquid.builtin import capture_tag
from liquid.builtin import cycle_tag
from liquid.builtin import assign_tag
from liquid.builtin import increment_tag
from liquid.builtin import decrement_tag
from liquid.builtin import echo_tag
from liquid.builtin import include_tag
from liquid.builtin import render_tag
from liquid.builtin import ifchanged_tag


class Case(NamedTuple):
    description: str
    template: str
    expect: Union[str, List[str]]
    expect_statements: int = 1

    @property
    def expected(self):
        if isinstance(self.expect, str):
            return [self.expect]
        return self.expect


# pylint: disable=too-many-public-methods
class ParserTestCase(unittest.TestCase):
    def _test(self, test_cases: Iterable[Case], instance, mode: Mode = Mode.STRICT):
        """Helper method for testing lists of `Case`s."""
        env = Environment()
        env.mode = mode

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)

                self.assertIsNotNone(template)
                self.assertEqual(len(template.tree.statements), case.expect_statements)
                if case.expect_statements:
                    self.assertIsInstance(template.tree.statements[0], instance)
                    self.assertIn(str(template.tree), case.expected)

    def test_template_literal(self):
        """Test that we can parse a template literal."""
        test_cases = [
            Case("plain html", "<HTML>Some text</HTML>", "<HTML>Some text</HTML>"),
        ]

        self._test(test_cases, literal.LiteralNode)

    def test_output_statement(self):
        """Test that we can parse output statements."""
        test_cases = [
            Case("string literal", r"{{ 'hello' }}", "`'hello'`"),
            Case("negative integer", r"{{ -5 }}", ["`(-5)`", "`-5`"]),
            Case("identifier", r"{{ product }}", "`product`"),
            Case(
                "identifier with a filter with args and kwargs",
                r"{{ product.title | default: 'foo', allow_false:true }}",
                "`product.title | default: 'foo', allow_false: True`",
            ),
        ]

        self._test(test_cases, statement.StatementNode)

    def test_illegal_tag(self):
        """Test that we can handle illegal tokens in lax mode."""

        test_cases = [
            Case(
                "unknown tag name",
                "{% foobar some %}{% endfoobar %}",
                "",
                expect_statements=0,
            ),
            Case(
                "unknown tag name with surrounding literal",
                "<HTML>{% foobar %}{% endfoobar %}</HTML>",
                "<HTML></HTML>",
                expect_statements=2,
            ),
        ]

        env = Environment(tolerance=Mode.LAX)

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)

                self.assertIsNotNone(template)
                self.assertEqual(len(template.tree.statements), case.expect_statements)

                if case.expect_statements:
                    self.assertEqual(str(template.tree), case.expect)

    def test_if_tag(self):
        """Test that we can parse an if expression."""
        test_cases = [
            Case(
                "condition with literal consequence",
                "{% if product == 'foo' %}foo{% endif %}",
                "if (product == 'foo') { foo }",
            ),
            Case(
                "condition with literal consequence and following literal",
                "{% if product == 'foo' %}foo{% endif %} hello",
                "if (product == 'foo') { foo } hello",
                expect_statements=2,
            ),
            Case(
                "condition with literal consequence and literal alternative",
                "{% if product == 'foo' %}foo{% else %}bar{% endif %}",
                "if (product == 'foo') { foo } else { bar }",
            ),
            Case(
                "condition with literal and object consequence",
                "{% if product == 'foo' %}foo {{collection}} bar{% endif %}",
                "if (product == 'foo') { foo `collection` bar }",
            ),
            Case(
                "condition with nested condition in the consequence block",
                (
                    "{% if product == 'foo' %}"
                    "{% if title == 'baz' %}baz{% endif %}"
                    "{% endif %}"
                ),
                "if (product == 'foo') { if (title == 'baz') { baz } }",
            ),
            Case(
                "condition with nested, alternative condition in the consequence block",
                (
                    "{% if product == 'foo' %}{% if title == 'baz' %}baz"
                    "{% else %}bar{% endif %}{% endif %}"
                ),
                "if (product == 'foo') { if (title == 'baz') { baz } else { bar } }",
            ),
            Case(
                "condition with nested condition in the alternative block",
                (
                    "{% if product == 'foo' %}"
                    "foo"
                    "{% else %}"
                    "{% if title == 'baz' %}bar{% endif %}"
                    "{% endif %}"
                ),
                "if (product == 'foo') { foo } else { if (title == 'baz') { bar } }",
            ),
            Case(
                "condition with conditional alternative",
                (
                    "{% if product == 'foo' %}"
                    "foo"
                    "{% elsif product == 'bar' %}"
                    "bar"
                    "{% endif %}"
                ),
                "if (product == 'foo') { foo } elsif (product == 'bar') { bar }",
            ),
            Case(
                "condition with conditional and final alternatives",
                (
                    "{% if product == 'foo' %}"
                    "foo"
                    "{% elsif product == 'bar' %}"
                    "bar"
                    "{% else %}"
                    "baz"
                    "{% endif %}"
                ),
                (
                    "if (product == 'foo') { foo } "
                    "elsif (product == 'bar') { bar } else { baz }"
                ),
            ),
            Case(
                "condition with multiple alternatives",
                (
                    "{% if product == 'foo' %}"
                    "foo"
                    "{% elsif product == 'bar' %}"
                    "bar{% elsif x == 1 %}"
                    "baz"
                    "{% endif %}"
                ),
                (
                    "if (product == 'foo') { foo } "
                    "elsif (product == 'bar') { bar } elsif (x == 1) { baz }"
                ),
            ),
        ]

        self._test(test_cases, if_tag.IfNode)

    def test_default_comment_tag(self):
        """Test that we can parse comment tags."""
        test_cases = [
            Case("simple comment", "{% comment %}foo{% endcomment %}", "/* foo */"),
            Case(
                "commented object",
                "{% comment %}{{ product }}{% endcomment %}",
                "/* product */",
            ),
        ]

        self._test(test_cases, comment_tag.CommentNode)

    def test_base_comment_tag(self):
        """Test that we can parse the space efficient base comment tag."""
        test_cases = [
            Case("simple comment", "{% comment %}foo{% endcomment %}", "/* */"),
            Case(
                "commented object",
                "{% comment %}{{ product }}{% endcomment %}",
                "/* */",
            ),
        ]

        env = Environment()
        env.add_tag(comment_tag.CommentTag)

        for case in test_cases:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                self.assertIn(str(template.tree), case.expected)

    def test_inline_comment_tag(self):
        """Test that we can parse inline comment tags."""
        test_cases = [
            Case("simple comment", "{% # foo %}", "/* foo */"),
            Case(
                "commented object",
                "{% # {{ product }} %}",
                "/* {{ product }} */",
            ),
        ]

        self._test(test_cases, comment_tag.CommentNode)

    def test_unless_tag(self):
        """Test that we can parse `unless` tags."""
        test_cases = [
            Case(
                "simple unless",
                "{% unless x == 1 %}foo{% endunless %}",
                "if !(x == 1) { foo }",
            ),
            Case(
                "unless with object and literal",
                "{% unless x == 1 %}foo {{ title }} bar{% endunless %}",
                "if !(x == 1) { foo `title` bar }",
            ),
            Case(
                "unless with alternative",
                "{% unless true %}foo{% else %}bar{% endunless %}",
                "if !(True) { foo } else { bar }",
            ),
            Case(
                "unless with conditional alternative",
                "{% unless true %}foo{% elsif true %}hello{% else %}bar{% endunless %}",
                "if !(True) { foo } elsif (True) { hello } else { bar }",
            ),
            Case(
                "unless with falsy conditional alternative",
                (
                    r"{% unless true %}foo"
                    r"{% elsif false %}bar"
                    r"{% else %}hello"
                    r"{% endunless %}"
                ),
                "if !(True) { foo } elsif (False) { bar } else { hello }",
            ),
        ]

        self._test(test_cases, unless_tag.UnlessNode)

    def test_case_tag(self):
        """Test that we can parse `case`/`when` tags."""
        test_cases = [
            Case(
                "simple case",
                "{% case title %}{% when 'foo' %}foo{% when 'bar' %}bar{% endcase %}",
                "if (title == 'foo') { foo } elsif (title == 'bar') { bar }",
            ),
            Case(
                "case with default",
                "{% case title %}{% when 'foo' %}foo{% else %}bar{% endcase %}",
                "if (title == 'foo') { foo } else { bar }",
            ),
            Case(
                "case with no whens",
                "{% case title %}{% else %}bar{% endcase %}",
                "if (False) { } else { bar }",
            ),
            Case(
                "case with no whens or default",
                "{% case title %}{% endcase %}",
                "if (False) { }",
            ),
            Case(
                "simple case with whitespace",
                "{% case title %} {% when 'foo' %}foo{% when 'bar' %}bar{% endcase %}",
                "if (title == 'foo') { foo } elsif (title == 'bar') { bar }",
            ),
        ]

        self._test(test_cases, case_tag.CaseNode)

    def test_for_tag(self):
        """Test that we can parse for loop tags."""
        test_cases = [
            Case(
                "simple for loop",
                "{% for prod in collection %}{{ prod.title }}{% endfor %}",
                "for (prod in collection) { `prod.title` }",
            ),
            Case(
                "for loop with default",
                "{% for prod in collection %}{{ prod.title }}{% else %}foo{% endfor %}",
                "for (prod in collection) { `prod.title` } else { foo }",
            ),
            Case(
                "for loop with break",
                (
                    "{% for prod in collection %}"
                    "{% if prod.title %}{{ prod.title }}{% else %}{% break %}"
                    "{% endif %}"
                    "{% endfor %}"
                ),
                (
                    "for (prod in collection) { "
                    "if (prod.title) { `prod.title` } else { `break` } "
                    "}"
                ),
            ),
            Case(
                "for loop with continue",
                (
                    "{% for prod in collection %}"
                    "{% if prod.title %}{{ prod.title }}"
                    "{% else %}{% continue %}"
                    "{% endif %}"
                    "{% endfor %}"
                ),
                (
                    "for (prod in collection) { "
                    "if (prod.title) { `prod.title` } else { `continue` } "
                    "}"
                ),
            ),
            Case(
                "for loop with limit",
                (
                    "{% for prod in collection limit:2 %}"
                    "{% if prod.title %}{{ prod.title }}"
                    "{% else %}{% continue %}"
                    "{% endif %}"
                    "{% endfor %}"
                ),
                (
                    "for (prod in collection limit:2) { "
                    "if (prod.title) { `prod.title` } else { `continue` } "
                    "}"
                ),
            ),
            Case(
                "for loop with offset",
                (
                    "{% for prod in collection offset:1 %}"
                    "{% if prod.title %}{{ prod.title }}"
                    "{% else %}{% continue %}"
                    "{% endif %}"
                    "{% endfor %}"
                ),
                (
                    "for (prod in collection offset:1) { "
                    "if (prod.title) { `prod.title` } else { `continue` } "
                    "}"
                ),
            ),
        ]

        self._test(test_cases, for_tag.ForNode)

    def test_tablerow_tag(self):
        """Test that we can parse tablerow tags."""
        test_cases = [
            Case(
                "simple table row",
                (
                    "{% tablerow prod in collection.products %}"
                    "{{ prod.title }}"
                    "{% endtablerow %}"
                ),
                "tablerow(prod in collection.products) { `prod.title` }",
            ),
            Case(
                "table row with whitespace",
                (
                    "{% tablerow prod in collection.products %} "
                    "{{ prod.title }} "
                    "{% endtablerow %}"
                ),
                "tablerow(prod in collection.products) {  `prod.title`  }",
            ),
            Case(
                "table row with arguments",
                (
                    "{% tablerow prod in collection.products cols:2 limit:3 %}"
                    "{{ prod.title }}"
                    "{% endtablerow %}"
                ),
                "tablerow(prod in collection.products limit:3 cols:2) { `prod.title` }",
            ),
        ]

        self._test(test_cases, tablerow_tag.TablerowNode)

    def test_raw_tag(self):
        """Test that we can parse raw tags."""
        test_cases = [
            Case("raw literal", "{% raw %}foo{% endraw %}", "foo"),
            Case("raw object", "{% raw %}{{ foo }}{% endraw %}", r"{{ foo }}"),
            Case(
                "raw tag",
                "{% raw %}{% assign x = 1 %}{% endraw %}",
                "{% assign x = 1 %}",
            ),
        ]

        self._test(test_cases, literal.LiteralNode)

    def test_capture_tag(self):
        """Test that we can parse capture tags."""
        test_cases = [
            Case(
                "capture literal",
                "{% capture some %}foo{% endcapture %}",
                "var some = { foo }",
            ),
            Case(
                "capture literal and statement",
                (
                    "{% capture greeting %}"
                    "Hello, {{ customer.first_name }}."
                    "{% endcapture %}"
                ),
                "var greeting = { Hello, `customer.first_name`. }",
            ),
        ]

        self._test(test_cases, capture_tag.CaptureNode)

    def test_cycle_tag(self):
        """Test that we can parse cycle tags."""
        test_cases = [
            Case(
                "cycle with no identifier",
                '{% cycle "some", "other" %}',
                "cycle ('some', 'other')",
            ),
            Case(
                "cycle with identifier",
                '{% cycle "foo": "some", "other" %}',
                "cycle ('foo': 'some', 'other')",
            ),
        ]

        self._test(test_cases, cycle_tag.CycleNode)

    def test_assign_tag(self):
        """Test that we can parse assign tags."""
        test_cases = [
            Case(
                "assign a string",
                '{% assign foo = "some" %}',
                "var (foo = 'some')",
            ),
        ]

        self._test(test_cases, assign_tag.AssignNode)

    def test_increment_tag(self):
        """Test that we can parse increment tags."""
        test_cases = [
            Case(
                "increment a named counter",
                "{% increment foo %}",
                "foo += 1",
            ),
        ]

        self._test(test_cases, increment_tag.IncrementNode)

    def test_decrement_tag(self):
        """Test that we can parse decrement tags."""
        test_cases = [
            Case(
                "decrement a named counter",
                "{% decrement foo %}",
                "foo -= 1",
            ),
        ]

        self._test(test_cases, decrement_tag.DecrementNode)

    def test_echo_tag(self):
        """Test that we can parse 'echo' tags."""
        test_cases = [
            Case("string literal", "{% echo 'hello' %}", "`'hello'`"),
            Case("identifier", "{% echo product %}", "`product`"),
            Case(
                "identifier with a filter",
                "{% echo product.title | upcase %}",
                "`product.title | upcase`",
            ),
        ]

        self._test(test_cases, echo_tag.EchoNode)

    def test_include_tag(self):
        """Test that we can parse 'include' tags."""
        test_cases = [
            Case(
                "simple string literal", "{% include 'product' %}", "include('product')"
            ),
            Case(
                "name from identifier", "{% include some.name %}", "include(some.name)"
            ),
            Case(
                "string literal with variable",
                "{% include 'product' with products[0] %}",
                "include('product' with products.0)",
            ),
            Case(
                "string literal with aliased variable",
                "{% include 'product' with products[0] as foo %}",
                "include('product' with products.0 as foo)",
            ),
            Case(
                "string literal with keyword argument",
                "{% include 'product', foo: 'bar' %}",
                "include('product', foo='bar')",
            ),
        ]

        self._test(test_cases, include_tag.IncludeNode)

    def test_render_tag(self):
        """Test that we can parse 'render' tags."""
        test_cases = [
            Case(
                "simple string literal", "{% render 'product' %}", "render('product')"
            ),
            Case(
                "string literal with variable",
                "{% render 'product' with products[0] %}",
                "render('product' with products.0)",
            ),
            Case(
                "string literal with aliased variable",
                "{% render 'product' with products[0] as foo %}",
                "render('product' with products.0 as foo)",
            ),
            Case(
                "string literal with keyword argument",
                "{% render 'product', foo: 'bar' %}",
                "render('product', foo='bar')",
            ),
        ]

        self._test(test_cases, render_tag.RenderNode)

    def test_ifchanged_tag(self):
        """Test that we can parse ifchanged tags."""
        test_cases = [
            Case(
                "initial state - no change",
                "{% ifchanged %}hello{% endifchanged %}",
                r"ifchanged { hello }",
            ),
        ]

        self._test(test_cases, ifchanged_tag.IfChangedNode)



"""Node and tag definitions for `with`."""
# pylint: disable=missing-class-docstring
from __future__ import annotations

import sys

from typing import TYPE_CHECKING
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import TextIO

from liquid.ast import ChildNode
from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.context import Context
from liquid.expression import Expression
from liquid.expressions import parse_keyword_arguments

from liquid.parse import expect
from liquid.parse import get_parser

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_EOF


if TYPE_CHECKING:  # pragma: no cover
    from liquid import Environment

TAG_WITH = sys.intern("with")
TAG_ENDWITH = sys.intern("endwith")


class WithKeywordArg(NamedTuple):
    name: str
    expr: Expression


class WithNode(Node):
    def __init__(self, tok: Token, args: Dict[str, Expression], block: BlockNode):
        self.tok = tok
        self.args = args
        self.block = block

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        namespace = {k: v.evaluate(context) for k, v in self.args.items()}

        with context.extend(namespace):
            return self.block.render(context, buffer)

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        namespace = {k: await v.evaluate_async(context) for k, v in self.args.items()}
        with context.extend(namespace):
            return await self.block.render_async(context, buffer)

    def children(self) -> List[ChildNode]:
        return [
            ChildNode(
                linenum=self.tok.linenum, node=self.block, block_scope=list(self.args)
            ),
            *[
                ChildNode(linenum=self.tok.linenum, expression=expr)
                for expr in self.args.values()
            ],
        ]


class WithTag(Tag):
    name = TAG_WITH
    end = TAG_ENDWITH

    def __init__(self, env: Environment):
        super().__init__(env)
        self.parser = get_parser(self.env)

    def parse(self, stream: TokenStream) -> Node:
        expect(stream, TOKEN_TAG, value=TAG_WITH)
        tok = next(stream)

        # Parse keyword arguments
        expect(stream, TOKEN_EXPRESSION)
        args = parse_keyword_arguments(stream.current.value)
        stream.next_token()

        # Parse the block
        block = self.parser.parse_block(stream, (TAG_ENDWITH, TOKEN_EOF))
        expect(stream, TOKEN_TAG, value=TAG_ENDWITH)

        return WithNode(tok=tok, args=args, block=block)



"""Parse tree node and tag definition for the built-in "capture" tag."""

import re
import sys

from io import StringIO

from typing import List
from typing import Optional
from typing import TextIO

try:
    from markupsafe import Markup
except ImportError:
    from liquid.exceptions import Markup  # type: ignore

from liquid import ast
from liquid.context import Context
from liquid.exceptions import LiquidSyntaxError

from liquid.parse import expect
from liquid.parse import get_parser

from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EOF
from liquid.token import TOKEN_EXPRESSION

from liquid.stream import TokenStream


RE_CAPTURE = re.compile(r"^\w[a-zA-Z0-9_\-]*$")

TAG_CAPTURE = sys.intern("capture")
TAG_ENDCAPTURE = sys.intern("endcapture")

ENDCAPTUREBLOCK = frozenset((TAG_ENDCAPTURE, TOKEN_EOF))


class CaptureNode(ast.Node):
    """Parse tree node for the built-in "capture" tag."""

    __slots__ = ("tok", "name", "block")

    def __init__(self, tok: Token, name: str, block: ast.BlockNode):
        self.tok = tok
        self.name = name
        self.block = block

    def __str__(self) -> str:
        return f"var {self.name} = {{ {self.block} }}"

    def __repr__(self) -> str:  # pragma: no cover
        return f"CaptureNode(tok={self.tok}, name={self.name}, block='{self.block}')"

    def _assign(self, context: Context, buf: StringIO) -> None:
        if context.autoescape:
            context.assign(self.name, Markup(buf.getvalue()))
        else:
            context.assign(self.name, buf.getvalue())

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        buf = context.get_buffer(buffer)
        self.block.render(context, buf)
        self._assign(context, buf)
        return False

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        buf = context.get_buffer(buffer)
        await self.block.render_async(context, buf)
        self._assign(context, buf)
        return False

    def children(self) -> List[ast.ChildNode]:
        return [
            ast.ChildNode(
                linenum=self.tok.linenum,
                node=self.block,
                template_scope=[self.name],
            )
        ]


class CaptureTag(Tag):
    """The built-in capture tag."""

    name = TAG_CAPTURE
    end = TAG_ENDCAPTURE

    def parse(self, stream: TokenStream) -> CaptureNode:
        parser = get_parser(self.env)

        expect(stream, TOKEN_TAG, value=TAG_CAPTURE)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)

        match = RE_CAPTURE.match(stream.current.value)
        if match:
            name = match.group()
        else:
            raise LiquidSyntaxError(
                f'invalid capture identifier "{stream.current.value}"',
                linenum=stream.current.linenum,
            )

        stream.next_token()
        block = parser.parse_block(stream, ENDCAPTUREBLOCK)
        expect(stream, TOKEN_TAG, value=TAG_ENDCAPTURE)

        return CaptureNode(tok, name, block=block)



"""Test built-in "undefined" types."""
import asyncio

from dataclasses import dataclass
from dataclasses import field

from unittest import TestCase
from typing import Dict

from liquid import Environment
from liquid import Undefined
from liquid import DebugUndefined
from liquid import StrictUndefined
from liquid import StrictDefaultUndefined

from liquid.template import BoundTemplate

from liquid.exceptions import UndefinedError
from liquid.exceptions import NoSuchFilterFunc


@dataclass
class Case:
    """Table driven test case helper."""

    description: str
    template: str
    expect: str
    context: Dict[str, object] = field(default_factory=dict)


class TestUndefined(TestCase):
    """Undefined test case."""

    def test_default_undefined(self):
        """Test that the default undefined type is quiet and forgiving."""
        tests = [
            Case(
                description="undefined in output statement",
                template=r"{{ nosuchthing }}",
                expect="",
            ),
            Case(
                description="undefined in loop expression",
                template=r"{% for tag in nosuchthing %}{tag}{% endfor %}",
                expect="",
            ),
            Case(
                description="index undefined",
                template=r"{{ nosuchthing[0] }}",
                expect="",
            ),
            Case(
                description="test undefined for truthy-ness",
                template=r"{% if nosuchthing %}hello{% endif %}",
                expect="",
            ),
            Case(
                description="compare undefined",
                template=r"{% if nosuchthing == 'hello' %}hello{% endif %}",
                expect="",
            ),
            Case(
                description="undefined equals undefined",
                template=r"{% if nosuchthing == noway %}hello{% endif %}",
                expect="hello",
            ),
            Case(
                description="undefined contains string",
                template=r"{% if nosuchthing contains 'hello' %}hello{% endif %}",
                expect="",
            ),
            Case(
                description="access `last` from undefined",
                template=r"{{ nosuchthing.last }}",
                expect="",
            ),
            Case(
                description="access `size` from undefined",
                template=r"{{ nosuchthing.size }}",
                expect="",
            ),
            Case(
                description="filtered undefined",
                template=r"hello {{ nosuchthing | last }} there",
                expect="hello  there",
            ),
            Case(
                description="math filter undefined",
                template=r"hello {{ nosuchthing | abs }} there",
                expect="hello 0 there",
            ),
            Case(
                description="undefined filter argument",
                template=r"hello {{ '1,2,3' | split: nosuchthing }} there",
                expect="hello 1,2,3 there",
            ),
            Case(
                description="filter undefined through date",
                template=r"hello {{ nosuchthing | date: '%b %d, %y' }} there",
                expect="hello  there",
            ),
            Case(
                description="array index out or range",
                template=r"{% assign a = '1,2,3,4,5' | split: ',' %}{{ a[100] }}",
                expect="",
            ),
            Case(
                description="negative array index out or range",
                template=r"{% assign a = '1,2,3,4,5' | split: ',' %}{{ a[-100] }}",
                expect="",
            ),
        ]

        env = Environment()

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render()
                self.assertEqual(case.expect, result)

    def test_strict_undefined(self):
        """Test that the strict undefined type raises an exception for everything."""
        tests = [
            Case(
                description="undefined in output statement",
                template=r"{{ nosuchthing }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined in loop expression",
                template=r"{% for tag in nosuchthing %}{tag}{% endfor %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="index undefined",
                template=r"{{ nosuchthing[0] }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="test undefined for truthy-ness",
                template=r"{% if nosuchthing %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="compare undefined",
                template=r"{% if nosuchthing == 'hello' %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined equals undefined",
                template=r"{% if nosuchthing == noway %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined contains string",
                template=r"{% if nosuchthing contains 'hello' %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="access `last` from undefined",
                template=r"{{ nosuchthing.last }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="access `size` from undefined",
                template=r"{{ nosuchthing.size }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="filtered undefined",
                template=r"hello {{ nosuchthing | last }} there",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined filter argument",
                template=r"hello {{ '1,2,3' | split: nosuchthing }} there",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="math filter undefined",
                template=r"hello {{ nosuchthing | abs }} there",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="array index out of range",
                template=r"{% assign a = '1,2,3,4,5' | split: ',' %}{{ a[100] }}",
                expect="list index out of range: a[100], on line 1",
            ),
            Case(
                description="negative array index out of range",
                template=r"{% assign a = '1,2,3,4,5' | split: ',' %}{{ a[-100] }}",
                expect="list index out of range: a[-100], on line 1",
            ),
            Case(
                description="key error",
                template=r"{{ obj['bar'] }}",
                expect="key error: 'bar', obj[bar], on line 1",
                context={"obj": {"foo": 1}},
            ),
            Case(
                description="default filter undefined",
                template=r"hello {{ nosuchthing | default: 'foo' }} there",
                expect="'nosuchthing' is undefined, on line 1",
            ),
        ]

        env = Environment(undefined=StrictUndefined)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)

                with self.assertRaises(UndefinedError) as raised:
                    template.render(**case.context)

                self.assertEqual(case.expect, str(raised.exception))

    def test_debug_undefined(self):
        """Test that the debugging undefined type prints debugging information."""
        env = Environment(undefined=DebugUndefined)
        template = env.from_string(r"{{ nosuchthing }}")
        result = template.render()
        self.assertEqual(result, "'nosuchthing' is undefined")

    def test_debug_undefined_hint(self):
        """Test that the debugging undefined type prints debugging hints."""
        env = Environment(undefined=DebugUndefined)
        undef = DebugUndefined(name="nosuchthing", hint="can't resolve identifier")
        template = env.from_string(r"{{ undef }}")
        result = template.render(undef=undef)
        self.assertEqual(result, "undefined: can't resolve identifier")

    def test_debug_undefined_object(self):
        """Test that the debugging undefined type prints related object information."""
        env = Environment(undefined=DebugUndefined)
        undef = DebugUndefined(name="nosuchthing", obj="foo")
        template = env.from_string(r"{{ nosuchthing }}")
        result = template.render(nosuchthing=undef)
        self.assertEqual(result, "str has no attribute 'nosuchthing'")

    def test_strict_default_undefined_with_default_filter(self):
        """Test that we can use an undefined type with the default filter."""
        env = Environment(undefined=StrictDefaultUndefined)
        template = env.from_string(r"{{ nosuchthing | default: 'hello' }}")
        result = template.render()
        self.assertEqual(result, "hello")

        template = env.from_string(r"{{ thing | default: 'hello' }}")
        result = template.render(thing="foo")
        self.assertEqual(result, "foo")

        template = env.from_string(r"{{ nosuchthing }}")
        with self.assertRaises(UndefinedError) as raised:
            template.render()

        self.assertEqual(str(raised.exception), "'nosuchthing' is undefined, on line 1")

    def test_filter_strict_default_undefined(self):
        """Test that the default undefined type raises an exception when used as a
        filter left value."""
        env = Environment(undefined=StrictDefaultUndefined)
        template = env.from_string(r"{{ nosuchthing | floor }}")
        with self.assertRaises(UndefinedError) as raised:
            template.render()

        self.assertEqual(str(raised.exception), "'nosuchthing' is undefined, on line 1")

    def test_isinstance_strict_default_filter(self):
        """Test that the default undefined type raises an exception when accessing
        __class__."""
        undef = StrictDefaultUndefined("nosuchthing")
        with self.assertRaises(UndefinedError):
            undef.__class__  # pylint: disable=pointless-statement

    def test_strict_default_undefined(self):
        """Test that the strict default undefined type raises an exception for
        everything other than the default filter."""
        tests = [
            Case(
                description="undefined in output statement",
                template=r"{{ nosuchthing }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined in loop expression",
                template=r"{% for tag in nosuchthing %}{tag}{% endfor %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="index undefined",
                template=r"{{ nosuchthing[0] }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="test undefined for truthy-ness",
                template=r"{% if nosuchthing %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="compare undefined",
                template=r"{% if nosuchthing == 'hello' %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined equals undefined",
                template=r"{% if nosuchthing == noway %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined contains string",
                template=r"{% if nosuchthing contains 'hello' %}hello{% endif %}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="access `last` from undefined",
                template=r"{{ nosuchthing.last }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="access `size` from undefined",
                template=r"{{ nosuchthing.size }}",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="filtered undefined",
                template=r"hello {{ nosuchthing | last }} there",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="undefined filter argument",
                template=r"hello {{ '1,2,3' | split: nosuchthing }} there",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="math filter undefined",
                template=r"hello {{ nosuchthing | abs }} there",
                expect="'nosuchthing' is undefined, on line 1",
            ),
            Case(
                description="array index out of range",
                template=r"{% assign a = '1,2,3,4,5' | split: ',' %}{{ a[100] }}",
                expect="list index out of range: a[100], on line 1",
            ),
            Case(
                description="negative array index out of range",
                template=r"{% assign a = '1,2,3,4,5' | split: ',' %}{{ a[-100] }}",
                expect="list index out of range: a[-100], on line 1",
            ),
            Case(
                description="key error",
                template=r"{{ obj['bar'] }}",
                expect="key error: 'bar', obj[bar], on line 1",
                context={"obj": {"foo": 1}},
            ),
        ]

        env = Environment(undefined=StrictDefaultUndefined)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)

                with self.assertRaises(UndefinedError) as raised:
                    template.render(**case.context)

                self.assertEqual(case.expect, str(raised.exception))

    def test_lax_filter(self):
        """Test that undefined filters can be silently ignored."""
        tests = [
            Case(
                description="undefined filter",
                template=r"{{ 'hello' | nosuchthing }}",
                expect="hello",
            ),
            Case(
                description="undefined filter with argument",
                template=r"{{ 'hello' | nosuchthing: 'foo' }}",
                expect="hello",
            ),
            Case(
                description="undefined filter with more filters",
                template=r"{{ 'hello' | nosuchthing | upcase }}",
                expect="HELLO",
            ),
        ]

        env = Environment(strict_filters=False)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)
                result = template.render()
                self.assertEqual(case.expect, result)

    def test_strict_filters(self):
        """Test that undefined filters raise an exception in strict mode."""
        tests = [
            Case(
                description="undefined filter",
                template=r"{{ 'hello' | nosuchthing }}",
                expect="unknown filter 'nosuchthing', on line 1",
            ),
            Case(
                description="undefined filter with argument",
                template=r"{{ 'hello' | nosuchthing: 'foo' }}",
                expect="unknown filter 'nosuchthing', on line 1",
            ),
            Case(
                description="undefined filter with more filters",
                template=r"{{ 'hello' | nosuchthing | upcase }}",
                expect="unknown filter 'nosuchthing', on line 1",
            ),
        ]

        env = Environment(strict_filters=True)

        for case in tests:
            with self.subTest(msg=case.description):
                template = env.from_string(case.template)

                with self.assertRaises(NoSuchFilterFunc) as raised:
                    template.render()

                self.assertEqual(case.expect, str(raised.exception))

                # And render async
                async def coro(template: BoundTemplate):
                    return await template.render_async()

                with self.assertRaises(NoSuchFilterFunc) as raised:
                    asyncio.run(coro(template))

    def test_default_undefined_magic(self):
        """Test the default undefined type magic methods."""
        undefined = Undefined("test")
        self.assertFalse("foo" in undefined)
        self.assertTrue(int(undefined) == 0)

    def test_strict_undefined_magic(self):
        """Test the strict undefined type magic methods."""
        undefined = StrictUndefined("test")

        with self.assertRaises(UndefinedError):
            self.assertFalse("foo" in undefined)

        with self.assertRaises(UndefinedError):
            int(undefined)  # type: ignore

        with self.assertRaises(UndefinedError):
            list(undefined)

        with self.assertRaises(UndefinedError):
            len(undefined)  # type: ignore

        with self.assertRaises(UndefinedError):
            str(undefined)

        self.assertEqual(repr(undefined), "StrictUndefined(test)")

        with self.assertRaises(UndefinedError):
            bool(undefined)

        with self.assertRaises(UndefinedError):
            hash(undefined)

        with self.assertRaises(UndefinedError):
            reversed(undefined)  # type: ignore



"""Mock currency filters."""
from liquid.filter import math_filter


@math_filter
def money_with_currency(money):
    """A filter function that returns a number formatted with currency info."""
    return f"$ {money / 100.0:.2f} USD"


@math_filter
def money_(money):
    """A filter function that returns a number formatted as money."""
    return f"$ {money / 100.0:.2f}"



"""Utilities for working with strings of HTML."""

from html.parser import HTMLParser

from typing import List
from typing import Tuple
from typing import Optional


class StripParser(HTMLParser):  # pylint: disable=abstract-method
    """An HTML parser that strips out tags."""

    def __init__(self) -> None:
        super().__init__(convert_charrefs=False)
        self.reset()
        self.script_depth = 0
        self.style_depth = 0
        self.dat: List[str] = []

    def handle_starttag(self, tag: str, attrs: List[Tuple[str, Optional[str]]]) -> None:
        if tag == "script":
            self.script_depth += 1
        elif tag == "style":
            self.style_depth += 1
        return super().handle_starttag(tag, attrs)

    def handle_endtag(self, tag: str) -> None:
        if tag == "script":
            self.script_depth -= 1
        elif tag == "style":
            self.style_depth -= 1
        return super().handle_endtag(tag)

    def handle_data(self, data: str) -> None:
        if not self.script_depth and not self.style_depth:
            self.dat.append(data)

    def handle_entityref(self, name: str) -> None:
        if not self.script_depth and not self.style_depth:
            self.dat.append(f"&{name};")

    def handle_charref(self, name: str) -> None:
        if not self.script_depth and not self.style_depth:
            self.dat.append(f"&#{name};")

    def get_data(self) -> str:
        """Return accumulated data."""
        return "".join(self.dat)


def strip_tags(value: str) -> str:
    """Return the given value with all HTML tags removed."""
    if "<" in value and ">" in value:
        parser = StripParser()
        parser.feed(value)
        parser.close()
        return parser.get_data()
    return value



"""Golden tests cases for testing liquid's built-in `sort` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="array of strings",
        template=r"{{ a | sort | join: '#' }}",
        expect="A#B#C#a#b",
        globals={"a": ["b", "a", "C", "B", "A"]},
    ),
    Case(
        description="array of objects",
        template=(
            r"{% assign x = a | sort: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,Baz)(title,bar)(title,foo)",
        globals={"a": [{"title": "foo"}, {"title": "Baz"}, {"title": "bar"}]},
    ),
    Case(
        description="array of objects with missing key",
        template=(
            r"{% assign x = a | sort: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,bar)(title,foo)(heading,Baz)",
        globals={"a": [{"title": "foo"}, {"heading": "Baz"}, {"title": "bar"}]},
    ),
    Case(
        description="empty array",
        template=r"{{ a | sort | join: '#' }}",
        expect="",
        globals={"a": []},
    ),
    Case(
        description="too many arguments",
        template=r"{{ a | sort: 'title', 'foo' | join: '#' }}",
        expect="",
        globals={"a": ["b", "a"]},
        error=True,
    ),
    Case(
        description="left value is not an array",
        template=r"{{ a | sort | join: '#' }}",
        expect="123",
        globals={"a": 123},
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | sort | join: '#' }}",
        expect="",
    ),
    Case(
        description="argument is undefined",
        template=r"{{ a | sort: nosuchthing | join: '#' }}",
        expect="a#b",
        globals={"a": ["b", "a"]},
    ),
    Case(
        description="sort a string",
        template=r"{{ 'BzAa4' | sort | join: '#' }}",
        expect="BzAa4",
    ),
    Case(
        description="array of integers",
        template=r"{{ a | sort | join: '#' }}",
        expect="1#3#30#1000",
        globals={"a": [1, 1000, 3, 30]},
    ),
    Case(
        description="incompatible types",
        template=r"{{ a | sort }}",
        expect="",
        globals={"a": [[], {}, 1, "4"]},
        error=True,
    ),
]



"""Golden tests cases for testing liquid's built-in `compact` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="array with a nil",
        template=r"{{ a | compact | join: '#' }}",
        expect="b#a#A",
        globals={"a": ["b", "a", None, "A"]},
    ),
    Case(
        description="empty array",
        template=r"{{ a | compact | join: '#' }}",
        expect="",
        globals={"a": []},
    ),
    Case(
        description="too many arguments",
        template=r"{{ a | compact: 'foo', 'bar' }}",
        expect="",
        error=True,
    ),
    Case(
        description="left value is not an array",
        template=r"{{ a | compact | first }}",
        expect="123",
        globals={"a": 123},
    ),
    Case(
        description="left value is undefined",
        template=r"{{ nosuchthing | compact }}",
        expect="",
    ),
    Case(
        description="array of objects with key property",
        template=(
            r"{% assign x = a | compact: 'title' %}"
            r"{% for obj in x %}"
            r"{% for i in obj %}"
            r"({{ i[0] }},{{ i[1] }})"
            r"{% endfor %}"
            r"{% endfor %}"
        ),
        expect="(title,foo)(name,a)(title,bar)(name,c)",
        globals={
            "a": [
                {"title": "foo", "name": "a"},
                {"title": None, "name": "b"},
                {"title": "bar", "name": "c"},
            ]
        },
    ),
]



"""Golden tests cases for testing liquid's built-in `escape` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="make HTML-safe",
        template=r'{{ "<p>test</p>" | escape }}',
        expect="&lt;p&gt;test&lt;/p&gt;",
    ),
    Case(
        description="not a string",
        template=r"{{ 5 | escape }}",
        expect="5",
    ),
    Case(
        description="unexpected argument",
        template=r'{{ "HELLO" | escape: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="undefined left value",
        template=r"{{ nosuchthing | escape }}",
        expect="",
    ),
]



"""Resource limit test cases."""
# pylint: disable=missing-class-docstring
import platform
import unittest

from liquid import DictLoader
from liquid import Environment

from liquid.exceptions import ContextDepthError
from liquid.exceptions import LoopIterationLimitError
from liquid.exceptions import LocalNamespaceLimitError
from liquid.exceptions import OutputStreamLimitError


class ContextDepthLimitTestCase(unittest.TestCase):
    """Render context depth limit test cases."""

    def test_recursive_render(self):
        """Test that we get an exception when rendering recursively."""
        env = Environment(
            loader=DictLoader(
                {
                    "foo": "{% render 'bar' %}",
                    "bar": "{% render 'foo' %}",
                }
            )
        )
        template = env.from_string("{% render 'foo' %}")
        with self.assertRaises(ContextDepthError):
            template.render()

    def test_recursive_include(self):
        """Test that we get an exception when including recursively."""
        env = Environment(
            loader=DictLoader(
                {
                    "foo": "{% include 'bar' %}",
                    "bar": "{% include 'foo' %}",
                }
            )
        )
        template = env.from_string("{% include 'foo' %}")
        with self.assertRaises(ContextDepthError):
            template.render()

    def test_set_context_depth_limit_render(self):
        """Test that we can set the context depth limit."""
        env = Environment(
            loader=DictLoader(
                {
                    "foo": "{% render 'bar' %}",
                    "bar": "{% render 'baz' %}",
                    "baz": "Hello",
                }
            )
        )
        template = env.from_string("{% render 'foo' %}")
        result = template.render()
        self.assertEqual(result, "Hello")

        env.context_depth_limit = 3
        template = env.from_string("{% render 'foo' %}")
        with self.assertRaises(ContextDepthError):
            template.render()

    def test_set_context_depth_limit_include(self):
        """Test that we can set the context depth limit."""
        env = Environment(
            loader=DictLoader(
                {
                    "foo": "{% include 'bar' %}",
                    "bar": "{% include 'baz' %}",
                    "baz": "Hello",
                }
            )
        )
        template = env.from_string("{% include 'foo' %}")
        result = template.render()
        self.assertEqual(result, "Hello")

        env.context_depth_limit = 3
        template = env.from_string("{% include 'foo' %}")
        with self.assertRaises(ContextDepthError):
            template.render()


class LoopIterationLimitTestCase(unittest.TestCase):
    """Loop iteration limit test cases."""

    def test_default_loop_iteration_limit(self):
        """Test (effectively) that the default loop limit is None."""
        env = Environment()
        template = env.from_string(
            "{% for i in (1..100) %}"
            "{% for j in (1..100) %}"
            "x"
            "{% endfor %}"
            "{% endfor %}"
        )

        template.render()

    def test_set_loop_iteration_limit(self):
        """Test that we can set the loop iteration limit."""

        class MockEnv(Environment):
            loop_iteration_limit = 10000

        env = MockEnv()
        env.from_string(
            "{% for i in (1..100) %}"
            "{% for j in (1..100) %}"
            "{{ i }},{{ j }}"
            "{% endfor %}"
            "{% endfor %}"
        ).render()

        template = env.from_string(
            "{% for i in (1..101) %}"
            "{% for j in (1..100) %}"
            "{{ i }},{{ j }}"
            "{% endfor %}"
            "{% endfor %}"
        )

        with self.assertRaises(LoopIterationLimitError):
            template.render()

    def test_render_carries_loop_count(self):
        """Test that a `render`ed template carries the loop count from its parent."""

        class MockEnv(Environment):
            loop_iteration_limit = 3000

        env = MockEnv(
            loader=DictLoader(
                {
                    "foo": (
                        "{% for i in (1..50) %}"
                        "{% for j in (1..50) %}"
                        "{{ i }},{{ j }}"
                        "{% endfor %}"
                        "{% endfor %}"
                    ),
                }
            )
        )
        template = env.from_string(
            "{% for i in (1..10) %}{% render 'foo' %}{% endfor %}"
        )

        with self.assertRaises(LoopIterationLimitError):
            template.render()

    def test_nested_renders_carry_loop_count(self):
        """Test that multiple, nested renders carry the loop count."""

        class MockEnv(Environment):
            loop_iteration_limit = 3000

        env = MockEnv(
            loader=DictLoader(
                {
                    "foo": (
                        "{% for i in (1..50) %}" "{% render 'bar' %}" "{% endfor %}"
                    ),
                    "bar": ("{% for j in (1..50) %}" "{{ j }}" "{% endfor %}"),
                }
            )
        )
        template = env.from_string(
            "{% for i in (1..10) %}{% render 'foo' %}{% endfor %}"
        )

        with self.assertRaises(LoopIterationLimitError):
            template.render()

    def test_include_contributes_to_count(self):
        """Test that `included`ed templates contribute to the loop count."""

        class MockEnv(Environment):
            loop_iteration_limit = 3000

        env = MockEnv(
            loader=DictLoader(
                {
                    "foo": (
                        "{% for i in (1..50) %}"
                        "{% for j in (1..50) %}"
                        "{{ i }},{{ j }}"
                        "{% endfor %}"
                        "{% endfor %}"
                    ),
                }
            )
        )
        template = env.from_string(
            "{% for i in (1..10) %}{% include 'foo' %}{% endfor %}"
        )

        with self.assertRaises(LoopIterationLimitError):
            template.render()

    def test_tablerow_contributes_to_count(self):
        """Test that the tablerow tag contributes to the loop count."""

        class MockEnv(Environment):
            loop_iteration_limit = 99

        env = MockEnv()
        template = env.from_string(
            "{% for i in (1..10) %}"
            "{% tablerow i in (1..10) cols:2 %}"
            "{{ i }}"
            "{% endtablerow %}"
            "{% endfor %}"
        )

        with self.assertRaises(LoopIterationLimitError):
            template.render()


class LocalNamespaceLimitTestCase(unittest.TestCase):
    """Render context local namespace length limit test cases."""

    @unittest.skipIf(platform.python_implementation() == "PyPy", "no sys.getsizeof")
    def test_set_local_namespace_limit(self):
        """Test that we can set the local namespace limit."""

        class MockEnv(Environment):
            local_namespace_limit = 140

        env = MockEnv()
        env.from_string(
            "{% assign a = 1 %}"
            "{% assign b = 2 %}"
            "{% assign c = 3 %}"
            "{% assign d = 4 %}"
            "{% assign e = 5 %}"
        ).render()

        template = env.from_string(
            "{% assign a = 1 %}"
            "{% assign b = 2 %}"
            "{% assign c = 3 %}"
            "{% assign d = 4 %}"
            "{% assign e = 5 %}"
            "{% assign f = 6 %}"
        )

        with self.assertRaises(LocalNamespaceLimitError):
            template.render()

    def test_copied_context_carries_parent_length(self):
        """Test that copied render context object cary the length of their parent
        context's local namespace."""

        class MockEnv(Environment):
            local_namespace_limit = 5

        env = MockEnv(
            loader=DictLoader(
                {
                    "foo": (
                        "{% assign a = 1 %}"
                        "{% assign b = 2 %}"
                        "{% assign c = 3 %}"
                        "{% assign d = 4 %}"
                        "{% assign e = 5 %}"
                        "{% assign e = 'five' %}"
                    )
                }
            )
        )

        template = env.from_string("{% assign f = 6 %}{% render 'foo' %}")

        with self.assertRaises(LocalNamespaceLimitError):
            template.render()

    def test_sizeof_local_namespace_with_unhashable_values(self):
        """Test that we can calculate the size of a local namespace when it contains
        unhashable objects."""

        class MockEnv(Environment):
            local_namespace_limit = 200

        env = MockEnv()
        env.from_string("{% assign foo = bar %}").render(bar=[1, 2, 3, 4])

        env.from_string(
            '{% assign beatles = "John, Paul, George, Ringo" | split: ", " %}'
        ).render()


class OutputStreamLimitTestCase(unittest.TestCase):
    """Template output stream limit test cases."""

    def test_set_output_stream_limit(self):
        """Test that we can set an output stream limit."""

        class MockEnv(Environment):
            output_stream_limit = 5

        env = MockEnv()
        env.from_string(
            "{% if false %}some literal that is longer then the limit{% endif %}"
            "hello"
        ).render()

        template = env.from_string(
            "{% if true %}some literal that is longer then the limit{% endif %}hello"
        )

        with self.assertRaises(OutputStreamLimitError):
            template.render()



"""Extra array filters."""
from typing import Sequence
from liquid.filter import array_filter


@array_filter
def index(left: Sequence[object], obj: object) -> object:
    """Return the zero-based index of an item in an array, or None if
    the items is not in the array.
    """
    try:
        return left.index(obj)
    except ValueError:
        return None



"""Tag an node definition for the built-in "inline comment" tag."""
import re
import sys

from liquid.builtin.tags.comment_tag import CommentNode
from liquid.exceptions import LiquidSyntaxError
from liquid.parse import expect
from liquid.tag import Tag
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION
from liquid.stream import TokenStream

RE_INVALID_INLINE_COMMENT = re.compile(r"\n\s*[^#\s]")


class InlineCommentTag(Tag):
    """The built-in inline comment ("#") tag."""

    block = False
    name = sys.intern("#")

    def parse(self, stream: TokenStream) -> CommentNode:
        expect(stream, TOKEN_TAG, value=self.name)
        tok = stream.current
        # Empty comment tag?
        if stream.peek.type == TOKEN_EXPRESSION:
            next(stream)
            if RE_INVALID_INLINE_COMMENT.search(stream.current.value):
                raise LiquidSyntaxError(
                    "every line of an inline comment must start with a '#' character",
                    linenum=stream.current.linenum,
                )
        return CommentNode(tok, text=stream.current.value)



"""Test include-style expression tokenization.

Include-style expressions are those found in the `include` tag,
the `render` tag, or any tag accepting comma separated, key/value
arguments.
"""

import unittest
from typing import Any
from typing import NamedTuple

from liquid.expressions.include import tokenize

from liquid.token import TOKEN_AS
from liquid.token import TOKEN_COLON
from liquid.token import TOKEN_COMMA
from liquid.token import TOKEN_DOT
from liquid.token import TOKEN_FOR
from liquid.token import TOKEN_IDENTIFIER
from liquid.token import TOKEN_IDENTINDEX
from liquid.token import TOKEN_INTEGER
from liquid.token import TOKEN_LPAREN
from liquid.token import TOKEN_RANGE
from liquid.token import TOKEN_RPAREN
from liquid.token import TOKEN_STRING
from liquid.token import TOKEN_WITH
from liquid.token import Token


class Case(NamedTuple):
    """Table-driven test helper."""

    description: str
    source: str
    expect: Any


class LexIncludeExpressionTestCase(unittest.TestCase):
    """Test cases for `include` expressions."""

    def test_lex_include_expression(self) -> None:
        """Test that we can tokenize `include` expressions."""
        test_cases = [
            Case(
                "string literal name no local variable",
                "'product'",
                [
                    Token(1, TOKEN_STRING, "product"),
                ],
            ),
            Case(
                "name from identifier and no local variable",
                "section.name",
                [
                    Token(1, TOKEN_IDENTIFIER, "section"),
                    Token(1, TOKEN_DOT, "."),
                    Token(1, TOKEN_IDENTIFIER, "name"),
                ],
            ),
            Case(
                "string literal name with identifier local variable",
                "'product' with products[0]",
                [
                    Token(1, TOKEN_STRING, "product"),
                    Token(1, TOKEN_WITH, "with"),
                    Token(1, TOKEN_IDENTIFIER, "products"),
                    Token(1, TOKEN_IDENTINDEX, "0"),
                ],
            ),
            Case(
                "string literal name with keyword arguments",
                "'product', foo: 'bar', some: other.tags",
                [
                    Token(1, TOKEN_STRING, "product"),
                    Token(1, TOKEN_COMMA, ","),
                    Token(1, TOKEN_IDENTIFIER, "foo"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_STRING, "bar"),
                    Token(1, TOKEN_COMMA, ","),
                    Token(1, TOKEN_IDENTIFIER, "some"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_IDENTIFIER, "other"),
                    Token(1, TOKEN_DOT, "."),
                    Token(1, TOKEN_IDENTIFIER, "tags"),
                ],
            ),
            Case(
                "string literal name with keyword arguments including a range literal",
                "'product', foo: (1..3)",
                [
                    Token(1, TOKEN_STRING, "product"),
                    Token(1, TOKEN_COMMA, ","),
                    Token(1, TOKEN_IDENTIFIER, "foo"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_LPAREN, "("),
                    Token(1, TOKEN_INTEGER, "1"),
                    Token(1, TOKEN_RANGE, ".."),
                    Token(1, TOKEN_INTEGER, "3"),
                    Token(1, TOKEN_RPAREN, ")"),
                ],
            ),
            Case(
                "string literal name with identifier aliased local variable",
                "'product' with products[0] as foo",
                [
                    Token(1, TOKEN_STRING, "product"),
                    Token(1, TOKEN_WITH, "with"),
                    Token(1, TOKEN_IDENTIFIER, "products"),
                    Token(1, TOKEN_IDENTINDEX, "0"),
                    Token(1, TOKEN_AS, "as"),
                    Token(1, TOKEN_IDENTIFIER, "foo"),
                ],
            ),
            Case(
                "string literal name with iterable local variable",
                "'product' for products",
                [
                    Token(1, TOKEN_STRING, "product"),
                    Token(1, TOKEN_FOR, "for"),
                    Token(1, TOKEN_IDENTIFIER, "products"),
                ],
            ),
            Case(
                "string literal name with aliased iterable local variable",
                "'product' for products as foo",
                [
                    Token(1, TOKEN_STRING, "product"),
                    Token(1, TOKEN_FOR, "for"),
                    Token(1, TOKEN_IDENTIFIER, "products"),
                    Token(1, TOKEN_AS, "as"),
                    Token(1, TOKEN_IDENTIFIER, "foo"),
                ],
            ),
            Case(
                "literal name with identifier aliased local variable and arguments",
                "'product' with products[0] as foo, bar: 42, baz: 'hello'",
                [
                    Token(1, TOKEN_STRING, "product"),
                    Token(1, TOKEN_WITH, "with"),
                    Token(1, TOKEN_IDENTIFIER, "products"),
                    Token(1, TOKEN_IDENTINDEX, "0"),
                    Token(1, TOKEN_AS, "as"),
                    Token(1, TOKEN_IDENTIFIER, "foo"),
                    Token(1, TOKEN_COMMA, ","),
                    Token(1, TOKEN_IDENTIFIER, "bar"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_INTEGER, "42"),
                    Token(1, TOKEN_COMMA, ","),
                    Token(1, TOKEN_IDENTIFIER, "baz"),
                    Token(1, TOKEN_COLON, ":"),
                    Token(1, TOKEN_STRING, "hello"),
                ],
            ),
        ]

        for case in test_cases:
            with self.subTest(msg=case.description):
                tokens = list(tokenize(case.source))
                self.assertEqual(len(tokens), len(case.expect))

                for got, want in zip(tokens, case.expect):
                    self.assertEqual(got, want)



"""Golden tests cases for testing liquid's `cycle` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="no identifier",
        template=(
            r"{% cycle 'some', 'other' %}"
            r"{% cycle 'some', 'other' %}"
            r"{% cycle 'some', 'other' %}"
        ),
        expect="someothersome",
    ),
    Case(
        description="with identifier",
        template=(
            r"{% cycle 'foo': 'some', 'other' %}"
            r"{% cycle 'some', 'other' %}"
            r"{% cycle 'foo': 'some', 'other' %}"
        ),
        expect="somesomeother",
    ),
    Case(
        description="different items",
        template=(
            r"{% cycle '1', '2', '3' %}"
            r"{% cycle '1', '2' %}"
            r"{% cycle '1', '2', '3' %}"
        ),
        expect="112",
    ),
    Case(
        description="integers",
        template=r"{% cycle 1, 2, 3 %}{% cycle 1, 2, 3 %}{% cycle 1, 2, 3 %}",
        expect="123",
    ),
    # Case(
    #     description="some global variables",
    #     template=r"{% cycle a, b, c %}{% cycle a, b, c %}{% cycle a, b, c %}",
    #     expect="123",
    #     globals={"a": 1, "b": 2, "c": 3},
    # ),
    Case(
        description="variable name",
        template=r"{% cycle a: 1, 2, 3 %}{% cycle a: 1, 2, 3 %}{% cycle a: 1, 2, 3 %}",
        expect="123",
        globals={"a": "foo"},
    ),
    Case(
        description="changing variable name",
        template=(
            r"{% cycle a: 1, 2, 3 %}"
            r"{% assign a = 'bar' %}"
            r"{% cycle a: 1, 2, 3 %}"
            r"{% cycle a: 1, 2, 3 %}"
        ),
        expect="112",
        globals={"a": "foo"},
    ),
    # Case(
    #     description="named with different items",
    #     template=(
    #         r"{% cycle 'a': 1, 2, 3 %}"
    #         r"{% cycle 'a': 7, 8, 9 %}"
    #         r"{% cycle 'a': 1, 2, 3 %}"
    #     ),
    #     expect="183",
    # ),
]



"""Golden tests cases for testing liquid's built-in `strip_newlines` filter."""

from liquid.golden.case import Case

cases = [
    Case(
        description="newline and other whitespace",
        template='{{ "hello there\nyou" | strip_newlines }}',
        expect="hello thereyou",
    ),
    Case(
        description="not a string",
        template="{{ 5 | strip_newlines }}",
        expect="5",
    ),
    Case(
        description="unexpected argument",
        template='{{ "hello" | strip_newlines: 5 }}',
        expect="",
        error=True,
    ),
    Case(
        description="reference implementation test 1",
        template='{{ "a\nb\nc" | strip_newlines }}',
        expect="abc",
    ),
    Case(
        description="reference implementation test 2",
        template='{{ "a\r\nb\nc" | strip_newlines }}',
        expect="abc",
    ),
    Case(
        description="undefined left value",
        template="{{ nosuchthing | strip_newlines }}",
        expect="",
    ),
]



"""Golden tests cases for testing liquid's `ifchanged` tag."""

from liquid.golden.case import Case

cases = [
    Case(
        description="changed from initial state",
        template=r"{% ifchanged %}hello{% endifchanged %}",
        expect="hello",
    ),
    Case(
        description="not changed from initial state",
        template=r"{% ifchanged %}{% endifchanged %}",
        expect="",
    ),
    Case(
        description="no change from assign",
        template=(
            r"{% assign foo = 'hello' %}"
            r"{% ifchanged %}{{ foo }}{% endifchanged %}"
            r"{% ifchanged %}{{ foo }}{% endifchanged %}"
        ),
        expect="hello",
    ),
    Case(
        description="change from assign",
        template=(
            r"{% assign foo = 'hello' %}"
            r"{% ifchanged %}{{ foo }}{% endifchanged %}"
            r"{% ifchanged %}{{ foo }}{% endifchanged %}"
            r"{% assign foo = 'goodbye' %}"
            r"{% ifchanged %}{{ foo }}{% endifchanged %}"
        ),
        expect="hellogoodbye",
    ),
    Case(
        description="within for loop",
        template=(
            r'{% assign list = "1,3,2,1,3,1,2" | split: "," | sort %}'
            r"{% for item in list -%}"
            r"{%- ifchanged %} {{ item }}{% endifchanged -%}"
            r"{%- endfor %}"
        ),
        expect=" 1 2 3",
    ),
]



"""Golden tests cases for testing illegal tags."""

from liquid.golden.case import Case

cases = [
    Case(
        description="unknown tag",
        template=r"{% nosuchthing %}",
        expect="",
        error=True,
        strict=True,
    ),
    Case(
        description="no addition operator",
        template=r"{% assign x = 1 + 2 %}{{ x }}",
        expect="",
        error=True,
        strict=True,
    ),
    Case(
        description="no subtraction operator",
        template=r"{% assign x = 1 - 2 %}{{ x }}",
        expect="",
        error=True,
        strict=True,
    ),
    Case(
        description="no multiplication operator",
        template=r"{% assign x = 2 %}{{ x * 3 }}",
        expect="",
        error=True,
        strict=True,
    ),
]



"""Liquid expression evaluator test cases."""
import unittest
from decimal import Decimal

from typing import Any
from typing import Mapping
from typing import NamedTuple

from liquid.environment import Environment
from liquid.context import Context
from liquid.stream import TokenStream

# Note: These are test cases for the legacy expression implementation.
# See `test_eval_*_expression.py` for expression evaluation test cases
# as used by all built-in tags.

from liquid.lex import tokenize_filtered_expression
from liquid.lex import tokenize_boolean_expression
from liquid.lex import tokenize_loop_expression
from liquid.lex import tokenize_assignment_expression

from liquid.parse import parse_filtered_expression
from liquid.parse import parse_boolean_expression
from liquid.parse import parse_assignment_expression
from liquid.parse import parse_loop_expression


class Case(NamedTuple):
    """Subtest helper"""

    description: str
    context: Mapping
    expression: str
    expect: Any


class LiquidStatementEvalTestCase(unittest.TestCase):
    """Liquid statement expression evaluator test cases."""

    def _test(self, test_cases, lex_func, parse_func):
        """Utility method for evaluating lists of test cases."""
        env = Environment()

        for case in test_cases:
            context = Context(env, case.context)
            with self.subTest(msg=case.description):
                tokens = TokenStream(lex_func(case.expression))
                expr = parse_func(tokens)
                res = expr.evaluate(context)
                self.assertEqual(res, case.expect)

    def test_eval_statement_expression(self):
        """Test that we can evaluate a liquid statement expression."""
        test_cases = [
            Case("string literal", {}, "'foobar'", "foobar"),
            Case("integer literal", {}, "7", 7),
            Case("float literal", {}, "7.5", 7.5),
            Case("negative integer literal", {}, "-7", -7),
            Case("negative float literal", {}, "-7.5", -7.5),
            Case(
                "single global object identifier",
                {"collection": "foo"},
                "collection",
                "foo",
            ),
            Case(
                "string literal with no arg filter",
                {},
                "'foo' | upcase",
                "FOO",
            ),
            Case(
                "object identifier with no arg filter",
                {"collection": {"title": "foo"}},
                "collection.title | upcase",
                "FOO",
            ),
            Case(
                "string literal with two arg filter",
                {},
                '"Liquid" | slice: 2, 5',
                "quid",
            ),
            Case(
                "string literal with two filters",
                {},
                '"Liquid" | slice: 2, 5 | upcase',
                "QUID",
            ),
            Case(
                "resolve identifier chain",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.c",
                "hello",
            ),
            Case(
                "resolve identifier chain not in context",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.x",
                None,
            ),
            Case(
                "resolve identifier chain containing whitespace.",
                {"a": {"b x": {"c": "hello", "array": [1, 2, 3]}}},
                "a['b x'].c",
                "hello",
            ),
            Case(
                "resolve identifier chain ending in an array",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.array",
                [1, 2, 3],
            ),
            Case(
                "resolve identifier chain ending in an array index",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.array.1",
                2,
            ),
            Case(
                "resolve identifier chain ending in an array index using subscript",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.array[1]",
                2,
            ),
            Case(
                "try to read past an array",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.array.foo",
                None,
            ),
            Case(
                "array `first` special method",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.array.first",
                1,
            ),
            Case(
                "array `last` special method",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.array.last",
                3,
            ),
            Case(
                "array `size` special method",
                {"a": {"b": {"c": "hello", "array": [1, 2, 3]}}},
                "a.b.array.size",
                3,
            ),
            Case(
                "size of an empty array",
                {"a": {"b": {"c": "hello", "array": []}}},
                "a.b.array.size",
                0,
            ),
            Case(
                "size of an object",
                {"a": {"b": {"c": "hello", "array": []}}},
                "a.b.size",
                2,
            ),
            Case(
                "nested and chained",
                {
                    "linklists": {"main": "main menu"},
                    "section": {"settings": {"menu": "main"}},
                },
                "linklists[section.settings.menu]",
                "main menu",
            ),
            Case(
                "array index using negative subscript",
                {"a": [1, 2, 3]},
                "a[-1]",
                3,
            ),
        ]

        self._test(test_cases, tokenize_filtered_expression, parse_filtered_expression)

    def test_eval_boolean_expression(self):
        """Test that we can evaluate boolean expressions."""
        test_cases = [
            Case(
                description="true literal",
                context={},
                expression="true",
                expect=True,
            ),
            Case(
                description="false literal",
                context={},
                expression="false",
                expect=False,
            ),
            Case(
                description="string literal",
                context={},
                expression="'some'",
                expect=True,
            ),
            Case(
                description="empty string",
                context={},
                expression="''",
                expect=True,
            ),
            Case(
                description="truthy identifier",
                context={"collection": {"title": "foo"}},
                expression="collection.title",
                expect=True,
            ),
            Case(
                description="falsey identifier",
                context={"collection": {"title": "foo"}},
                expression="collection.tags",
                expect=False,
            ),
            Case(
                description="truthy comparision",
                context={"user": {"age": 21}},
                expression="user.age >= 21",
                expect=True,
            ),
            Case(
                description="not equal comparision",
                context={"user": {"age": 21}},
                expression="user.age != 21",
                expect=False,
            ),
            Case(
                description="truthy comparision and logic operator",
                context={
                    "user": {"age": 20},
                    "collection": {
                        "tags": [
                            "safe",
                        ]
                    },
                },
                expression="user.age >= 21 or collection.tags contains 'safe'",
                expect=True,
            ),
            Case(
                description="boolean with logic operators",
                context={},
                expression="true and false and false or true",
                expect=False,
            ),
            Case(
                description="empty array",
                context={"a": {"array": []}},
                expression="a.array == empty",
                expect=True,
            ),
            Case(
                description="empty object",
                context={"a": {"obj": {}}},
                expression="a.obj == empty",
                expect=True,
            ),
            Case(
                description="not empty array",
                context={"a": {"array": [1, 2]}},
                expression="a.array == empty",
                expect=False,
            ),
            Case(
                description="not empty object",
                context={"a": {"obj": {"foo": "bar"}}},
                expression="a.obj == empty",
                expect=False,
            ),
            Case(
                description="invalid comparison to empty",
                context={"a": {"foo": 1}},
                expression="a.foo == empty",
                expect=False,
            ),
            Case(
                description="empty equals empty",
                context={},
                expression="empty == empty",
                expect=True,
            ),
            Case(
                description="empty not equals true",
                context={},
                expression="empty != true",
                expect=True,
            ),
            Case(
                description="nil equals nil",
                context={},
                expression="nil == nil",
                expect=True,
            ),
            Case(
                description="nil equals null",
                context={},
                expression="nil == null",
                expect=True,
            ),
            Case(
                description="null equals null",
                context={},
                expression="null == null",
                expect=True,
            ),
            Case(
                description="string contains string",
                context={},
                expression="'hello' contains 'ell'",
                expect=True,
            ),
            Case(
                description="string contains int",
                context={},
                expression="'hel1lo' contains 1",
                expect=True,
            ),
            Case(
                description="string not equal int",
                context={},
                expression="'hello' != 1",
                expect=True,
            ),
            Case(
                description="array contains",
                context={"foo": [1, 2, 4]},
                expression="foo contains 2",
                expect=True,
            ),
            Case(
                description="array does not contain",
                context={"foo": [1, 2, 4]},
                expression="foo contains 3",
                expect=False,
            ),
            Case(
                description="int equals",
                context={},
                expression="1 == 1",
                expect=True,
            ),
            Case(
                description="int less than",
                context={},
                expression="1 < 2",
                expect=True,
            ),
            Case(
                description="int less than or equal",
                context={},
                expression="1 <= 1",
                expect=True,
            ),
            Case(
                description="int greater than",
                context={},
                expression="1 > 0",
                expect=True,
            ),
            Case(
                description="int greater than or equal",
                context={},
                expression="1 >= 1",
                expect=True,
            ),
            Case(
                description="true equals true",
                context={},
                expression="true == true",
                expect=True,
            ),
            Case(
                description="true equal false",
                context={},
                expression="true == false",
                expect=False,
            ),
            Case(
                description="true not equal false",
                context={},
                expression="true != false",
                expect=True,
            ),
            Case(
                description="string equals int",
                context={},
                expression="'2' == 2",
                expect=False,
            ),
            Case(
                description="empty string is truthy",
                context={},
                expression="''",
                expect=True,
            ),
            Case(
                description="empty string and string is truthy",
                context={},
                expression="'' and 'foo'",
                expect=True,
            ),
            Case(
                description="float equals int",
                context={},
                expression="1 == 1.0",
                expect=True,
            ),
            Case(
                description="empty string equals blank",
                context={},
                expression="'' == blank",
                expect=True,
            ),
            Case(
                description="empty string equals empty",
                context={},
                expression="'' == empty",
                expect=True,
            ),
            Case(
                description="empty array equals empty",
                context={"x": []},
                expression="x == empty",
                expect=True,
            ),
            Case(
                description="empty array equals blank",
                context={"x": []},
                expression="x == blank",
                expect=True,
            ),
            Case(
                description="whitespace string equals blank",
                context={"x": "   "},
                expression="x == blank",
                expect=True,
            ),
            Case(
                description="whitespace string does not equal empty",
                context={"x": "   "},
                expression="x == empty",
                expect=False,
            ),
            Case(
                description="blank does not equal empty",
                context={},
                expression="blank == empty",
                expect=False,
            ),
            Case(
                description="empty does not equal blank",
                context={},
                expression="empty == blank",
                expect=False,
            ),
            Case(
                description="blank does not equal empty",
                context={},
                expression="blank != empty",
                expect=True,
            ),
            Case(
                description="undefined equals nil",
                context={},
                expression="nosuchthing == nil",
                expect=True,
            ),
            Case(
                description="undefined equals null",
                context={},
                expression="nosuchthing == null",
                expect=True,
            ),
            Case(
                description="nil equals undefined",
                context={},
                expression="nil == nosuchthing",
                expect=True,
            ),
            Case(
                description="nil equals None",
                context={"nothing": None},
                expression="nil == nothing",
                expect=True,
            ),
            Case(
                description="nil equals nil",
                context={},
                expression="nil == nil",
                expect=True,
            ),
            Case(
                description="zero",
                context={},
                expression="0",
                expect=True,
            ),
            Case(
                description="0.0",
                context={},
                expression="0.0",
                expect=True,
            ),
            Case(
                description="one",
                context={},
                expression="1",
                expect=True,
            ),
            Case(
                description="zero equals false",
                context={},
                expression="0 == false",
                expect=False,
            ),
            Case(
                description="zero equals true",
                context={},
                expression="0 == true",
                expect=False,
            ),
            Case(
                description="0.0 equals false",
                context={},
                expression="0.0 == false",
                expect=False,
            ),
            Case(
                description="0.0 equals true",
                context={},
                expression="0.0 == true",
                expect=False,
            ),
            Case(
                description="one equals true",
                context={},
                expression="1 == true",
                expect=False,
            ),
            Case(
                description="true equals one",
                context={},
                expression="true == 1",
                expect=False,
            ),
            Case(
                description="one is less than true",
                context={},
                expression="1 < true",
                expect=False,
            ),
            Case(
                description="zero is less than true",
                context={},
                expression="0 < true",
                expect=False,
            ),
            Case(
                description="0.0 is less than true",
                context={},
                expression="0.0 < true",
                expect=False,
            ),
            Case(
                description="one is not equal true",
                context={},
                expression="1 != true",
                expect=True,
            ),
            Case(
                description="zero is not equal false",
                context={},
                expression="0 != false",
                expect=True,
            ),
            Case(
                description="0.0 is not equal false",
                context={},
                expression="0.0 != false",
                expect=True,
            ),
            Case(
                description="false is not equal zero",
                context={},
                expression="false != 0",
                expect=True,
            ),
            Case(
                description="false is not equal 0.0",
                context={},
                expression="false != 0.0",
                expect=True,
            ),
            Case(
                description="false is less than string",
                context={},
                expression="false < 'false'",
                expect=False,
            ),
            Case(
                description="decimal zero",
                context={"n": Decimal("0")},
                expression="n",
                expect=True,
            ),
            Case(
                description="decimal non-zero",
                context={"n": Decimal("1")},
                expression="n",
                expect=True,
            ),
            Case(
                description="decimal zero equals false",
                context={"n": Decimal("0")},
                expression="n == false",
                expect=False,
            ),
            Case(
                description="decimal zero equals true",
                context={"n": Decimal("0")},
                expression="n == true",
                expect=False,
            ),
        ]

        self._test(test_cases, tokenize_boolean_expression, parse_boolean_expression)

    def test_eval_assignment_expression(self):
        """Test that we can evaluate assignment expressions."""

        test_cases = [
            Case(
                description="assign a string literal",
                context={},
                expression="some = 'foo'",
                expect=("some", "foo"),
            ),
            Case(
                description="assign an integer literal",
                context={},
                expression="some = 5",
                expect=("some", 5),
            ),
            Case(
                description="assign a float literal",
                context={},
                expression="some = 5.7",
                expect=("some", 5.7),
            ),
            Case(
                description="assign an array using a split filter",
                context={},
                expression='some = "apples, oranges, peaches" | split: ", "',
                expect=("some", ["apples", "oranges", "peaches"]),
            ),
            Case(
                description="assign from a filtered identifier",
                context={"user": {"title": "Mr"}},
                expression="title = user.title | downcase",
                expect=("title", "mr"),
            ),
        ]

        env = Environment()

        for case in test_cases:
            context = Context(env, case.context)
            with self.subTest(msg=case.description):
                stream = TokenStream(tokenize_assignment_expression(case.expression))
                expr = parse_assignment_expression(stream)
                expr.evaluate(context)

                name, val = case.expect
                self.assertEqual(context.get(name), val)

    def test_eval_loop_expression(self):
        """Test that we can evaluate loop expressions."""

        test_cases = [
            Case(
                description="simple range loop",
                context={},
                expression="i in (0..3)",
                expect=[0, 1, 2, 3],
            ),
            Case(
                description="reversed range",
                context={},
                expression="i in (0..3) reversed",
                expect=[3, 2, 1, 0],
            ),
            Case(
                description="loop over an object from context",
                context={"a": {"name": "foo", "title": "bar"}},
                expression="i in a",
                expect=[("name", "foo"), ("title", "bar")],
            ),
            Case(
                description="loop over nested and chained object from context",
                context={
                    "linklists": {"main": ["1", "2"]},
                    "section": {"settings": {"menu": "main"}},
                },
                expression="link in linklists[section.settings.menu]",
                expect=["1", "2"],
            ),
            Case(
                description=(
                    "loop over nested and chained object from context "
                    "with trailing identifier"
                ),
                context={
                    "linklists": {"main": {"links": ["1", "2"]}},
                    "section": {"settings": {"menu": "main"}},
                },
                expression="link in linklists[section.settings.menu].links",
                expect=["1", "2"],
            ),
            Case(
                description="chained identifier in range loop",
                context={
                    "foo": {"bar": 3},
                },
                expression="x in (1..foo.bar)",
                expect=[1, 2, 3],
            ),
        ]

        env = Environment()

        for case in test_cases:
            context = Context(env, case.context)
            with self.subTest(msg=case.description):
                stream = TokenStream(tokenize_loop_expression(case.expression))
                expr = parse_loop_expression(stream)
                loopiter, length = expr.evaluate(context)

                self.assertEqual(list(loopiter), case.expect)
                self.assertEqual(length, len(case.expect))

    def test_eval_continue_loop_expression(self):
        """Test that we can evaluate loop expressions that use a continue offset."""

        env = Environment()
        context = Context(env, {"array": [1, 2, 3, 4, 5, 6]})

        # Mock a for loop with a limit
        context.stopindex("item-array", 3)

        stream = TokenStream(tokenize_loop_expression("item in array offset: continue"))
        expr = parse_loop_expression(stream)

        loopiter, length = expr.evaluate(context)

        self.assertEqual(list(loopiter), [4, 5, 6])
        self.assertEqual(length, 3)



"""Tag and node definition for the built-in "for" tag."""
from __future__ import annotations
import sys

from typing import List
from typing import Optional
from typing import Any
from typing import Mapping
from typing import TextIO
from typing import Iterator

from liquid.ast import ChildNode
from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.context import Context
from liquid.expression import LoopExpression

from liquid.exceptions import BreakLoop
from liquid.exceptions import ContinueLoop

from liquid.parse import get_parser
from liquid.parse import expect

from liquid.stream import TokenStream
from liquid.tag import Tag

from liquid.token import Token
from liquid.token import TOKEN_EXPRESSION
from liquid.token import TOKEN_TAG


TAG_FOR = sys.intern("for")
TAG_ENDFOR = sys.intern("endfor")
TAG_ELSE = sys.intern("else")

TAG_BREAK = sys.intern("break")
TAG_CONTINUE = sys.intern("continue")

ENDFORBLOCK = frozenset((TAG_ENDFOR, TAG_ELSE))
ENDFORELSEBLOCK = frozenset((TAG_ENDFOR,))


# pylint: disable=too-many-instance-attributes
class ForLoop(Mapping[str, object]):
    """Loop helper variables."""

    __slots__ = (
        "name",
        "it",
        "length",
        "item",
        "_index",
        "parentloop",
    )

    _keys = frozenset(
        [
            "name",
            "length",
            "index",
            "index0",
            "rindex",
            "rindex0",
            "first",
            "last",
            "parentloop",
        ]
    )

    def __init__(
        self,
        name: str,
        it: Iterator[Any],
        length: int,
        parentloop: object,
    ):
        self.name = name
        self.it = it
        self.length = length

        self.item = None
        self._index = -1  # Step is called before `next(it)`
        self.parentloop = parentloop

    def __repr__(self) -> str:  # pragma: no cover
        return f"ForLoop(name='{self.name}', length={self.length})"

    def __getitem__(self, key: str) -> object:
        if key in self._keys:
            return getattr(self, key)
        raise KeyError(key)

    def __len__(self) -> int:
        return len(self._keys)

    def __next__(self) -> object:
        self.step()
        return next(self.it)

    def __iter__(self) -> Iterator[Any]:
        return self

    def __str__(self) -> str:
        return "ForLoop"

    @property
    def index(self) -> int:
        """The 1-based index of the current loop iteration."""
        return self._index + 1

    @property
    def index0(self) -> int:
        """The 0-based index of the current loop iteration."""
        return self._index

    @property
    def rindex(self) -> int:
        """The 1-based index, counting from the right, of the current loop iteration."""
        return self.length - self._index

    @property
    def rindex0(self) -> int:
        """The 0-based index, counting from the right, of the current loop iteration."""
        return self.length - self._index - 1

    @property
    def first(self) -> bool:
        """True if this is the first iteration, false otherwise."""
        return self._index == 0

    @property
    def last(self) -> bool:
        """True if this is the last iteration, false otherwise."""
        return self._index == self.length - 1

    def step(self) -> None:
        """Move the for loop helper forward to the next iteration."""
        self._index += 1


class ForNode(Node):
    """Parse tree node for the built-in "for" tag."""

    __slots__ = ("tok", "expression", "block", "default", "forced_output")

    def __init__(
        self,
        tok: Token,
        expression: LoopExpression,
        block: BlockNode,
        default: Optional[BlockNode] = None,
    ):
        self.tok = tok
        self.expression = expression
        self.block = block
        self.default = default

        self.forced_output = any(
            b.forced_output for b in (self.block, self.default) if b
        )

    def __str__(self) -> str:
        tag_str = f"for ({self.expression}) {{ {self.block} }}"

        if self.default:
            tag_str += f" else {{ {self.default} }}"

        return tag_str

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        # This intermediate buffer is used to detect and possibly suppress blocks that,
        # when rendered, contain only whitespace.
        buf = context.get_buffer(buffer)
        rendered: Optional[bool] = False

        loop_iter, length = self.expression.evaluate(context)

        if length:
            rendered = True
            name = self.expression.name

            forloop = ForLoop(
                name=f"{name}-{self.expression.iterable}",
                it=loop_iter,
                length=length,
                parentloop=context.parentloop(),
            )

            namespace = {
                "forloop": forloop,
                name: None,
            }

            # Extend the context. Essentially giving priority to `ForLoopDrop`, then
            # delegating `get` and `assign` to the outer context.
            with context.loop(namespace, forloop):
                for itm in forloop:
                    namespace[name] = itm
                    try:
                        self.block.render(context, buf)
                    except ContinueLoop:
                        continue
                    except BreakLoop:
                        break

        elif self.default:
            rendered = self.default.render(context, buf)

        val = buf.getvalue()
        if self.forced_output or not val.isspace():
            buffer.write(val)

        return rendered

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        buf = context.get_buffer(buffer)
        rendered: Optional[bool] = False

        loop_iter, length = await self.expression.evaluate_async(context)

        if length:
            rendered = True
            name = self.expression.name

            forloop = ForLoop(
                name=f"{name}-{self.expression.iterable}",
                it=loop_iter,
                length=length,
                parentloop=context.parentloop(),
            )

            namespace = {
                "forloop": forloop,
                name: None,
            }

            # Extend the context. Essentially giving priority to `ForLoopDrop`, then
            # delegating `get` and `assign` to the outer context.
            with context.loop(namespace, forloop):
                for itm in forloop:
                    namespace[name] = itm
                    try:
                        await self.block.render_async(context, buf)
                    except ContinueLoop:
                        continue
                    except BreakLoop:
                        break

        elif self.default:
            rendered = await self.default.render_async(context, buf)

        val = buf.getvalue()
        if self.forced_output or not val.isspace():
            buffer.write(val)

        return rendered

    def children(self) -> List[ChildNode]:
        _children = [
            ChildNode(
                linenum=self.block.tok.linenum,
                node=self.block,
                expression=self.expression,
                block_scope=[self.expression.name, "forloop"],
            )
        ]
        if self.default:
            _children.append(
                ChildNode(
                    linenum=self.default.tok.linenum,
                    node=self.default,
                )
            )
        return _children


class BreakNode(Node):
    """Parse tree node for the built-in "break" tag."""

    __slots__ = ("tok",)

    def __init__(self, tok: Token):
        self.tok = tok

    def __str__(self) -> str:
        return "`break`"

    def render_to_output(
        self,
        context: Context,
        buffer: TextIO,
    ) -> Optional[bool]:
        raise BreakLoop("break")


class ContinueNode(Node):
    """Parse tree node for the built-in "continue" tag."""

    __slots__ = ("tok",)

    def __init__(self, tok: Token):
        self.tok = tok

    def __str__(self) -> str:
        return "`continue`"

    def render_to_output(
        self,
        context: Context,
        buffer: TextIO,
    ) -> Optional[bool]:
        raise ContinueLoop("continue")


class ForTag(Tag):
    """The built-in "for" tag."""

    name = TAG_FOR
    end = TAG_ENDFOR

    def parse(self, stream: TokenStream) -> Node:
        parser = get_parser(self.env)

        expect(stream, TOKEN_TAG, value=TAG_FOR)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        expr = self.env.parse_loop_expression_value(stream.current.value)
        stream.next_token()

        block = parser.parse_block(stream, ENDFORBLOCK)
        default: Optional[BlockNode] = None

        if stream.current.istag(TAG_ELSE):
            stream.next_token()
            default = parser.parse_block(stream, ENDFORELSEBLOCK)

        expect(stream, TOKEN_TAG, value=TAG_ENDFOR)
        return ForNode(tok, expression=expr, block=block, default=default)


class BreakTag(Tag):
    """The built-in "break" tag."""

    name = TAG_BREAK
    block = False

    def parse(self, stream: TokenStream) -> BreakNode:
        expect(stream, TOKEN_TAG, value=TAG_BREAK)
        return BreakNode(stream.current)


class ContinueTag(Tag):
    """The built-in "continue" tag."""

    name = TAG_CONTINUE
    block = False

    def parse(self, stream: TokenStream) -> ContinueNode:
        expect(stream, TOKEN_TAG, value=TAG_CONTINUE)
        return ContinueNode(stream.current)



"""Tag and node definition for the built-in "tablerow" tag."""
import sys

from typing import Any
from typing import Dict
from typing import List
from typing import Mapping
from typing import Optional
from typing import TextIO
from typing import Iterator

from liquid.token import Token
from liquid.token import TOKEN_TAG
from liquid.token import TOKEN_EXPRESSION

from liquid.ast import ChildNode
from liquid.ast import Node
from liquid.ast import BlockNode

from liquid.context import Context
from liquid.expression import LoopExpression
from liquid.expression import NIL
from liquid.limits import to_int

from liquid.parse import expect
from liquid.parse import get_parser

from liquid.tag import Tag
from liquid.stream import TokenStream


TAG_TABLEROW = sys.intern("tablerow")
TAG_ENDTABLEROW = sys.intern("endtablerow")

END_TAGBLOCK = frozenset((TAG_ENDTABLEROW,))


# pylint: disable=too-many-instance-attributes
class TableRow(Mapping[str, object]):
    """Table row helper variables."""

    __slots__ = (
        "name",
        "it",
        "length",
        "ncols",
        "_index",
        "_row",
        "_col",
    )

    _keys = frozenset(
        [
            "length",
            "index",
            "index0",
            "rindex",
            "rindex0",
            "first",
            "last",
            "col",
            "col0",
            "col_first",
            "col_last",
            "row",
        ]
    )

    def __init__(self, name: str, it: Iterator[Any], length: int, ncols: int) -> None:
        self.name = name
        self.it = it
        self.length = length
        self.ncols = ncols
        self._index = -1
        self._row = 1
        self._col = 0

    def __repr__(self) -> str:  # pragma: no cover
        return f"TableRow(name='{self.name}', length={self.length})"

    def __getitem__(self, key: str) -> object:
        if key in self._keys:
            return getattr(self, key)
        raise KeyError(key)

    def __len__(self) -> int:
        return len(self._keys)

    def __iter__(self) -> Iterator[Any]:
        return self

    def __next__(self) -> object:
        self.step()
        return next(self.it)

    @property
    def index(self) -> int:
        """The 1-based index of the current loop iteration."""
        return self._index + 1

    @property
    def index0(self) -> int:
        """The 0-based index of the current loop iteration."""
        return self._index

    @property
    def rindex(self) -> int:
        """The 1-based index, counting from the right, of the current loop iteration."""
        return self.length - self._index

    @property
    def rindex0(self) -> int:
        """The 0-based index, counting from the right, of the current loop iteration."""
        return self.length - self._index - 1

    @property
    def first(self) -> bool:
        """True if this is the first iteration, false otherwise."""
        return self._index == 0

    @property
    def last(self) -> bool:
        """True if this is the last iteration, false otherwise."""
        return self._index == self.length - 1

    @property
    def col(self) -> int:
        """The 1-based index of the current column."""
        return self._col

    @property
    def col0(self) -> int:
        """The 0-based index of the current column."""
        return self._col - 1

    @property
    def col_first(self) -> bool:
        """True if this is the first column. False otherwise"""
        return self._col == 1

    @property
    def col_last(self) -> bool:
        """True if this is the last iteration, false otherwise."""
        return self._col == self.ncols

    @property
    def row(self) -> int:
        """The current row number."""
        return self._row

    def step(self) -> None:
        """Step the forloop forward."""
        self._index += 1
        if self._col == self.ncols:
            self._col = 1
            self._row += 1
        else:
            self._col += 1


class TablerowNode(Node):
    """Parse tree node for the built-in "tablerow" tag."""

    __slots__ = ("tok", "expression", "block")

    def __init__(
        self,
        tok: Token,
        expression: LoopExpression,
        block: BlockNode,
    ):
        self.tok = tok
        self.expression = expression
        self.block = block

    def __str__(self) -> str:
        return f"tablerow({ self.expression }) {{ {self.block} }}"

    def _int_or_zero(self, arg: object) -> int:
        try:
            return to_int(arg)
        except ValueError:
            return 0

    def render_to_output(self, context: Context, buffer: TextIO) -> Optional[bool]:
        name = self.expression.name
        loop_iter, length = self.expression.evaluate(context)

        if self.expression.cols and self.expression.cols != NIL:
            cols = self._int_or_zero(self.expression.cols.evaluate(context))
        else:
            cols = length

        context.raise_for_loop_limit(length)
        tablerow = TableRow(name, loop_iter, length, cols)

        namespace: Dict[str, object] = {
            "tablerowloop": tablerow,
        }

        buffer.write('<tr class="row1">\n')

        with context.extend(namespace):
            for item in tablerow:
                namespace[name] = item
                buffer.write(f'<td class="col{tablerow.col}">')
                self.block.render(context=context, buffer=buffer)
                buffer.write("</td>")

                if tablerow.col_last and not tablerow.last:
                    buffer.write(f'</tr>\n<tr class="row{tablerow.row + 1}">')

            buffer.write("</tr>\n")
        return True

    async def render_to_output_async(
        self, context: Context, buffer: TextIO
    ) -> Optional[bool]:
        name = self.expression.name
        loop_iter, length = await self.expression.evaluate_async(context)

        if self.expression.cols and self.expression.cols != NIL:
            cols = self._int_or_zero(await self.expression.cols.evaluate_async(context))
        else:
            cols = length

        context.raise_for_loop_limit(length)
        tablerow = TableRow(name, loop_iter, length, cols)

        namespace: Dict[str, object] = {
            "tablerowloop": tablerow,
        }

        buffer.write('<tr class="row1">\n')

        with context.extend(namespace):
            for item in tablerow:
                namespace[name] = item
                buffer.write(f'<td class="col{tablerow.col}">')
                await self.block.render_async(context=context, buffer=buffer)
                buffer.write("</td>")

                if tablerow.col_last and not tablerow.last:
                    buffer.write(f'</tr>\n<tr class="row{tablerow.row + 1}">')

            buffer.write("</tr>\n")
        return True

    def children(self) -> List[ChildNode]:
        return [
            ChildNode(
                linenum=self.block.tok.linenum,
                node=self.block,
                expression=self.expression,
                block_scope=["tablerowloop", self.expression.name],
            )
        ]


class TablerowTag(Tag):
    """The built-in "tablerow" tag."""

    name = TAG_TABLEROW
    end = TAG_ENDTABLEROW

    def parse(self, stream: TokenStream) -> TablerowNode:
        parser = get_parser(self.env)

        expect(stream, TOKEN_TAG, value=TAG_TABLEROW)
        tok = stream.current
        stream.next_token()

        expect(stream, TOKEN_EXPRESSION)
        loop_expression = self.env.parse_loop_expression_value(stream.current.value)
        stream.next_token()

        block = parser.parse_block(stream, END_TAGBLOCK)
        expect(stream, TOKEN_TAG, value=TAG_ENDTABLEROW)

        return TablerowNode(tok, expression=loop_expression, block=block)
